import json
import logging
import secrets
import threading

import pykka
import tornado.httpserver
import tornado.ioloop
import tornado.netutil
import tornado.web
import tornado.websocket

from mopidy import exceptions, models, zeroconf
from mopidy.core import CoreListener
from mopidy.http import Extension, handlers
from mopidy.internal import formatting, network

try:
    import asyncio
except ImportError:
    asyncio = None


logger = logging.getLogger(__name__)


class HttpFrontend(pykka.ThreadingActor, CoreListener):
    apps = []
    statics = []

    def __init__(self, config, core):
        super().__init__()

        self.hostname = network.format_hostname(config["http"]["hostname"])
        self.port = config["http"]["port"]
        tornado_hostname = config["http"]["hostname"]
        if tornado_hostname == "::":
            tornado_hostname = None

        try:
            logger.debug("Starting HTTP server")
            sockets = tornado.netutil.bind_sockets(self.port, tornado_hostname)
            self.server = HttpServer(
                config=config,
                core=core,
                sockets=sockets,
                apps=self.apps,
                statics=self.statics,
            )
        except OSError as exc:
            raise exceptions.FrontendError(f"HTTP server startup failed: {exc}")

        self.zeroconf_name = config["http"]["zeroconf"]
        self.zeroconf_http = None
        self.zeroconf_mopidy_http = None

    def on_start(self):
        logger.info("HTTP server running at [%s]:%s", self.hostname, self.port)
        self.server.start()

        if self.zeroconf_name:
            self.zeroconf_http = zeroconf.Zeroconf(
                name=self.zeroconf_name, stype="_http._tcp", port=self.port
            )
            self.zeroconf_mopidy_http = zeroconf.Zeroconf(
                name=self.zeroconf_name,
                stype="_mopidy-http._tcp",
                port=self.port,
            )
            self.zeroconf_http.publish()
            self.zeroconf_mopidy_http.publish()

    def on_stop(self):
        if self.zeroconf_http:
            self.zeroconf_http.unpublish()
        if self.zeroconf_mopidy_http:
            self.zeroconf_mopidy_http.unpublish()

        self.server.stop()

    def on_event(self, name, **data):
        on_event(name, self.server.io_loop, **data)


def on_event(name, io_loop, **data):
    event = data
    event["event"] = name
    message = json.dumps(event, cls=models.ModelJSONEncoder)
    handlers.WebSocketHandler.broadcast(message, io_loop)


class HttpServer(threading.Thread):
    name = "HttpServer"

    def __init__(self, config, core, sockets, apps, statics):
        super().__init__()

        self.config = config
        self.core = core
        self.sockets = sockets
        self.apps = apps
        self.statics = statics

        self.app = None
        self.server = None
        self.io_loop = None

    def run(self):
        if asyncio:
            # If asyncio is available, Tornado uses it as its IO loop. Since we
            # start Tornado in a another thread than the main thread, we must
            # explicitly create an asyncio loop for the current thread.
            asyncio.set_event_loop(asyncio.new_event_loop())

        self.app = tornado.web.Application(
            self._get_request_handlers(),
            cookie_secret=self._get_cookie_secret(),
        )
        self.server = tornado.httpserver.HTTPServer(self.app)
        self.server.add_sockets(self.sockets)

        self.io_loop = tornado.ioloop.IOLoop.current()
        self.io_loop.start()

        logger.debug("Stopped HTTP server")

    def stop(self):
        logger.debug("Stopping HTTP server")
        self.io_loop.add_callback(self.io_loop.stop)

    def _get_request_handlers(self):
        request_handlers = []
        request_handlers.extend(self._get_app_request_handlers())
        request_handlers.extend(self._get_static_request_handlers())
        request_handlers.extend(self._get_default_request_handlers())

        logger.debug(
            "HTTP routes from extensions: %s",
            formatting.indent(
                "\n".join(
                    f"{path!r}: {handler!r}"
                    for (path, handler, *_) in request_handlers
                )
            ),
        )

        return request_handlers

    def _get_app_request_handlers(self):
        result = []
        for app in self.apps:
            try:
                request_handlers = app["factory"](self.config, self.core)
            except Exception:
                logger.exception("Loading %s failed.", app["name"])
                continue

            result.append((f"/{app['name']}", handlers.AddSlashHandler))
            for handler in request_handlers:
                handler = list(handler)
                handler[0] = f"/{app['name']}{handler[0]}"
                result.append(tuple(handler))
            logger.debug("Loaded HTTP extension: %s", app["name"])
        return result

    def _get_static_request_handlers(self):
        result = []
        for static in self.statics:
            result.append((f"/{static['name']}", handlers.AddSlashHandler))
            result.append(
                (
                    f"/{static['name']}/(.*)",
                    handlers.StaticFileHandler,
                    {"path": static["path"], "default_filename": "index.html"},
                )
            )
            logger.debug("Loaded static HTTP extension: %s", static["name"])
        return result

    def _get_default_request_handlers(self):
        sites = [app["name"] for app in self.apps + self.statics]

        default_app = self.config["http"]["default_app"]
        if default_app not in sites:
            logger.warning(
                f"HTTP server's default app {default_app!r} not found"
            )
            default_app = "mopidy"
        logger.debug(f"Default webclient is {default_app}")

        return [
            (
                r"/",
                tornado.web.RedirectHandler,
                {"url": f"/{default_app}/", "permanent": False},
            )
        ]

    def _get_cookie_secret(self):
        file_path = Extension.get_data_dir(self.config) / "cookie_secret"
        if not file_path.is_file():
            cookie_secret = secrets.token_hex(32)
            file_path.write_text(cookie_secret)
        else:
            cookie_secret = file_path.read_text().strip()
            if not cookie_secret:
                logging.error(
                    f"HTTP server could not find cookie secret in {file_path}"
                )
        return cookie_secret



import os
import subprocess

import mopidy


def get_version():
    try:
        return get_git_version()
    except OSError:
        return mopidy.__version__


def get_git_version():
    project_dir = os.path.abspath(
        os.path.join(os.path.dirname(mopidy.__file__), "..")
    )
    process = subprocess.Popen(
        ["git", "describe"],
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
        cwd=project_dir,
    )
    if process.wait() != 0:
        raise OSError('Execution of "git describe" failed')
    version = process.stdout.read().strip().decode()
    if version.startswith("v"):
        version = version[1:]
    return version



from pytest import raises

from mopidy import exceptions
from mopidy.internal import validation


def test_check_boolean_with_valid_values():
    for value in (True, False):
        validation.check_boolean(value)


def test_check_boolean_with_other_values():
    for value in 1, 0, None, "", list(), tuple():
        with raises(exceptions.ValidationError):
            validation.check_boolean(value)


def test_check_boolean_error_message():
    with raises(exceptions.ValidationError) as excinfo:
        validation.check_boolean(1234)
    assert "Expected a boolean, not 1234" == str(excinfo.value)


def test_check_choice_with_valid_values():
    for value, choices in (2, (1, 2, 3)), ("abc", ("abc", "def")):
        validation.check_choice(value, choices)


def test_check_choice_with_invalid_values():
    for value, choices in (5, (1, 2, 3)), ("xyz", ("abc", "def")):
        with raises(exceptions.ValidationError):
            validation.check_choice(value, choices)


def test_check_choice_error_message():
    with raises(exceptions.ValidationError) as excinfo:
        validation.check_choice(5, (1, 2, 3))
    assert "Expected one of (1, 2, 3), not 5" == str(excinfo.value)


def test_check_instance_with_valid_choices():
    for value, cls in ((True, bool), ("a", str), (123, int)):
        validation.check_instance(value, cls)


def test_check_instance_with_invalid_values():
    for value, cls in (1, str), ("abc", int):
        with raises(exceptions.ValidationError):
            validation.check_instance(value, cls)


def test_check_instance_error_message():
    with raises(exceptions.ValidationError) as excinfo:
        validation.check_instance(1, dict)
    assert "Expected a dict instance, not 1" == str(excinfo.value)


def test_check_instances_with_valid_values():
    validation.check_instances([], int)
    validation.check_instances([1, 2], int)
    validation.check_instances((1, 2), int)


def test_check_instances_with_invalid_values():
    with raises(exceptions.ValidationError):
        validation.check_instances("abc", str)
    with raises(exceptions.ValidationError):
        validation.check_instances(["abc", 123], str)
    with raises(exceptions.ValidationError):
        validation.check_instances(None, str)
    with raises(exceptions.ValidationError):
        validation.check_instances([None], str)
    with raises(exceptions.ValidationError):
        validation.check_instances(iter(["abc"]), str)


def test_check_instances_error_message():
    with raises(exceptions.ValidationError) as excinfo:
        validation.check_instances([1], str)
    assert "Expected a list of str, not [1]" == str(excinfo.value)


def test_check_query_valid_values():
    for value in {}, {"any": []}, {"any": ["abc"]}:
        validation.check_query(value)


def test_check_query_random_iterables():
    for value in None, tuple(), list(), "abc":
        with raises(exceptions.ValidationError):
            validation.check_query(value)


def test_check_mapping_error_message():
    with raises(exceptions.ValidationError) as excinfo:
        validation.check_query([])
    assert "Expected a query dictionary, not []" == str(excinfo.value)


def test_check_query_invalid_fields():
    for value in "wrong", "bar", "foo", "tlid":
        with raises(exceptions.ValidationError):
            validation.check_query({value: []})


def test_check_field_error_message():
    with raises(exceptions.ValidationError) as excinfo:
        validation.check_query({"wrong": ["abc"]})
    assert "Expected query field to be one of " in str(excinfo.value)


def test_check_query_invalid_values():
    for value in "", None, "foo", 123, [""], [None], iter(["abc"]):
        with raises(exceptions.ValidationError):
            validation.check_query({"any": value})


def test_check_values_error_message():
    with raises(exceptions.ValidationError) as excinfo:
        validation.check_query({"any": "abc"})
    assert 'Expected "any" to be list of strings, not' in str(excinfo.value)


def test_check_uri_with_valid_values():
    for value in "foobar:", "http://example.com", "git+http://example.com":
        validation.check_uri(value)


def test_check_uri_with_invalid_values():
    # Note that tuple catches a potential bug with using "'foo' % arg" for
    # formatting.
    for value in ("foobar", "htt p://example.com", None, 1234, tuple()):
        with raises(exceptions.ValidationError):
            validation.check_uri(value)


def test_check_uri_error_message():
    with raises(exceptions.ValidationError) as excinfo:
        validation.check_uri("testing")
    assert "Expected a valid URI, not 'testing'" == str(excinfo.value)


def test_check_uris_with_valid_values():
    validation.check_uris([])
    validation.check_uris(["foobar:"])
    validation.check_uris(("foobar:",))


def test_check_uris_with_invalid_values():
    with raises(exceptions.ValidationError):
        validation.check_uris("foobar:")
    with raises(exceptions.ValidationError):
        validation.check_uris(None)
    with raises(exceptions.ValidationError):
        validation.check_uris([None])
    with raises(exceptions.ValidationError):
        validation.check_uris(["foobar:", "foobar"])
    with raises(exceptions.ValidationError):
        validation.check_uris(iter(["http://example.com"]))


def test_check_uris_error_message():
    with raises(exceptions.ValidationError) as excinfo:
        validation.check_uris("testing")
    assert "Expected a list of URIs, not 'testing'" == str(excinfo.value)



import threading
import unittest
from unittest import mock

import pykka

from mopidy import audio
from mopidy.audio.constants import PlaybackState
from mopidy.internal import path
from mopidy.internal.gi import Gst

from tests import dummy_audio, path_to_data_dir

# We want to make sure both our real audio class and the fake one behave
# correctly. So each test is first run against the real class, then repeated
# against our dummy.


class BaseTest(unittest.TestCase):
    config = {
        "audio": {
            "buffer_time": None,
            "mixer": "fakemixer track_max_volume=65536",
            "mixer_track": None,
            "mixer_volume": None,
            "output": "testoutput",
            "visualizer": None,
        }
    }

    uris = [
        path.path_to_uri(path_to_data_dir("song1.wav")),
        path.path_to_uri(path_to_data_dir("song2.wav")),
    ]

    audio_class = audio.Audio

    def setUp(self):  # noqa: N802
        config = {
            "audio": {
                "buffer_time": None,
                "mixer": "foomixer",
                "mixer_volume": None,
                "output": "testoutput",
                "visualizer": None,
            },
            "proxy": {"hostname": ""},
        }
        self.song_uri = path.path_to_uri(path_to_data_dir("song1.wav"))
        self.audio = self.audio_class.start(config=config, mixer=None).proxy()

    def tearDown(self):  # noqa
        pykka.ActorRegistry.stop_all()

    def possibly_trigger_fake_playback_error(self, uri):
        pass

    def possibly_trigger_fake_about_to_finish(self):
        pass


class DummyMixin:
    audio_class = dummy_audio.DummyAudio

    def possibly_trigger_fake_playback_error(self, uri):
        self.audio.trigger_fake_playback_failure(uri)

    def possibly_trigger_fake_about_to_finish(self):
        callback = self.audio.get_about_to_finish_callback().get()
        if callback:
            callback()


class AudioTest(BaseTest):
    def test_start_playback_existing_file(self):
        self.audio.prepare_change()
        self.audio.set_uri(self.uris[0])
        assert self.audio.start_playback().get()

    def test_start_playback_non_existing_file(self):
        self.possibly_trigger_fake_playback_error(self.uris[0] + "bogus")

        self.audio.prepare_change()
        self.audio.set_uri(self.uris[0] + "bogus")
        assert not self.audio.start_playback().get()

    def test_pause_playback_while_playing(self):
        self.audio.prepare_change()
        self.audio.set_uri(self.uris[0])
        self.audio.start_playback()
        assert self.audio.pause_playback().get()

    def test_stop_playback_while_playing(self):
        self.audio.prepare_change()
        self.audio.set_uri(self.uris[0])
        self.audio.start_playback()
        assert self.audio.stop_playback().get()

    @unittest.SkipTest
    def test_deliver_data(self):
        pass  # TODO

    @unittest.SkipTest
    def test_end_of_data_stream(self):
        pass  # TODO

    @unittest.SkipTest
    def test_set_mute(self):
        pass  # TODO Probably needs a fakemixer with a mixer track

    @unittest.SkipTest
    def test_set_state_encapsulation(self):
        pass  # TODO

    @unittest.SkipTest
    def test_set_position(self):
        pass  # TODO

    @unittest.SkipTest
    def test_invalid_output_raises_error(self):
        pass  # TODO


class AudioDummyTest(DummyMixin, AudioTest):
    pass


class DummyAudioListener(pykka.ThreadingActor, audio.AudioListener):
    def __init__(self):
        super().__init__()
        self.events = []
        self.waiters = {}

    def on_event(self, event, **kwargs):
        self.events.append((event, kwargs))
        if event in self.waiters:
            self.waiters[event].set()

    def wait(self, event):
        self.waiters[event] = threading.Event()
        return self.waiters[event]

    def get_events(self):
        return self.events

    def clear_events(self):
        self.events = []


class AudioEventTest(BaseTest):
    def setUp(self):  # noqa: N802
        super().setUp()
        self.audio.enable_sync_handler().get()
        self.listener = DummyAudioListener.start().proxy()

    def tearDown(self):  # noqa: N802
        super().tearDown()

    def assertEvent(self, event, **kwargs):  # noqa: N802
        assert (event, kwargs) in self.listener.get_events().get()

    def assertNotEvent(self, event, **kwargs):  # noqa: N802
        assert (event, kwargs) not in self.listener.get_events().get()

    # TODO: test without uri set, with bad uri and gapless...
    # TODO: playing->playing triggered by seek should be removed
    # TODO: codify expected state after EOS
    # TODO: consider returning a future or a threading event?

    def test_state_change_stopped_to_playing_event(self):
        self.audio.prepare_change()
        self.audio.set_uri(self.uris[0])
        self.audio.start_playback()

        self.audio.wait_for_state_change().get()
        self.assertEvent(
            "state_changed",
            old_state=PlaybackState.STOPPED,
            new_state=PlaybackState.PLAYING,
            target_state=None,
        )

    def test_state_change_stopped_to_paused_event(self):
        self.audio.prepare_change()
        self.audio.set_uri(self.uris[0])
        self.audio.pause_playback()

        self.audio.wait_for_state_change().get()
        self.assertEvent(
            "state_changed",
            old_state=PlaybackState.STOPPED,
            new_state=PlaybackState.PAUSED,
            target_state=None,
        )

    def test_state_change_paused_to_playing_event(self):
        self.audio.prepare_change()
        self.audio.set_uri(self.uris[0])
        self.audio.pause_playback()

        self.audio.wait_for_state_change()
        self.listener.clear_events()
        self.audio.start_playback()

        self.audio.wait_for_state_change().get()
        self.assertEvent(
            "state_changed",
            old_state=PlaybackState.PAUSED,
            new_state=PlaybackState.PLAYING,
            target_state=None,
        )

    def test_state_change_paused_to_stopped_event(self):
        self.audio.prepare_change()
        self.audio.set_uri(self.uris[0])
        self.audio.pause_playback()

        self.audio.wait_for_state_change()
        self.listener.clear_events()
        self.audio.stop_playback()

        self.audio.wait_for_state_change().get()
        self.assertEvent(
            "state_changed",
            old_state=PlaybackState.PAUSED,
            new_state=PlaybackState.STOPPED,
            target_state=None,
        )

    def test_state_change_playing_to_paused_event(self):
        self.audio.prepare_change()
        self.audio.set_uri(self.uris[0])
        self.audio.start_playback()

        self.audio.wait_for_state_change()
        self.listener.clear_events()
        self.audio.pause_playback()

        self.audio.wait_for_state_change().get()
        self.assertEvent(
            "state_changed",
            old_state=PlaybackState.PLAYING,
            new_state=PlaybackState.PAUSED,
            target_state=None,
        )

    def test_state_change_playing_to_stopped_event(self):
        self.audio.prepare_change()
        self.audio.set_uri(self.uris[0])
        self.audio.start_playback()

        self.audio.wait_for_state_change()
        self.listener.clear_events()
        self.audio.stop_playback()

        self.audio.wait_for_state_change().get()
        self.assertEvent(
            "state_changed",
            old_state=PlaybackState.PLAYING,
            new_state=PlaybackState.STOPPED,
            target_state=None,
        )

    def test_stream_changed_event_on_playing(self):
        self.audio.prepare_change()
        self.audio.set_uri(self.uris[0])
        self.listener.clear_events()
        self.audio.start_playback()

        # Since we are going from stopped to playing, the state change is
        # enough to ensure the stream changed.
        self.audio.wait_for_state_change().get()
        self.assertEvent("stream_changed", uri=self.uris[0])

    def test_stream_changed_event_on_multiple_changes(self):
        self.audio.prepare_change()
        self.audio.set_uri(self.uris[0])
        self.listener.clear_events()
        self.audio.start_playback()

        self.audio.wait_for_state_change().get()
        self.assertEvent("stream_changed", uri=self.uris[0])

        self.audio.prepare_change()
        self.audio.set_uri(self.uris[1])
        self.audio.pause_playback()

        self.audio.wait_for_state_change().get()
        self.assertEvent("stream_changed", uri=self.uris[1])

    def test_stream_changed_event_on_playing_to_paused(self):
        self.audio.prepare_change()
        self.audio.set_uri(self.uris[0])
        self.listener.clear_events()
        self.audio.start_playback()

        self.audio.wait_for_state_change().get()
        self.assertEvent("stream_changed", uri=self.uris[0])

        self.listener.clear_events()
        self.audio.pause_playback()

        self.audio.wait_for_state_change().get()
        self.assertNotEvent("stream_changed", uri=self.uris[0])

    def test_stream_changed_event_on_paused_to_stopped(self):
        self.audio.prepare_change()
        self.audio.set_uri(self.uris[0])
        self.audio.pause_playback()

        self.audio.wait_for_state_change()
        self.listener.clear_events()
        self.audio.stop_playback()

        self.audio.wait_for_state_change().get()
        self.assertEvent("stream_changed", uri=None)

    def test_position_changed_on_pause(self):
        self.audio.prepare_change()
        self.audio.set_uri(self.uris[0])
        self.audio.pause_playback()
        self.audio.wait_for_state_change()

        self.audio.wait_for_state_change().get()
        self.assertEvent("position_changed", position=0)

    def test_stream_changed_event_on_paused_to_playing(self):
        self.audio.prepare_change()
        self.audio.set_uri(self.uris[0])
        self.listener.clear_events()
        self.audio.pause_playback()

        self.audio.wait_for_state_change().get()
        self.assertEvent("stream_changed", uri=self.uris[0])

        self.listener.clear_events()
        self.audio.start_playback()

        self.audio.wait_for_state_change().get()
        self.assertNotEvent("stream_changed", uri=self.uris[0])

    def test_position_changed_on_play(self):
        self.audio.prepare_change()
        self.audio.set_uri(self.uris[0])
        self.audio.start_playback()
        self.audio.wait_for_state_change()

        self.audio.wait_for_state_change().get()
        self.assertEvent("position_changed", position=0)

    def test_position_changed_on_seek_while_stopped(self):
        self.audio.prepare_change()
        self.audio.set_uri(self.uris[0])
        self.audio.set_position(2000)

        self.audio.wait_for_state_change().get()
        self.assertNotEvent("position_changed", position=0)

    def test_position_changed_on_seek_after_play(self):
        self.audio.prepare_change()
        self.audio.set_uri(self.uris[0])
        self.audio.start_playback()

        self.audio.wait_for_state_change()
        self.listener.clear_events()
        self.audio.set_position(2000)

        self.audio.wait_for_state_change().get()
        self.assertEvent("position_changed", position=2000)

    def test_position_changed_on_seek_after_pause(self):
        self.audio.prepare_change()
        self.audio.set_uri(self.uris[0])
        self.audio.pause_playback()

        self.audio.wait_for_state_change()
        self.listener.clear_events()
        self.audio.set_position(2000)

        self.audio.wait_for_state_change().get()
        self.assertEvent("position_changed", position=2000)

    def test_tags_changed_on_playback(self):
        self.audio.prepare_change()
        self.audio.set_uri(self.uris[0])
        self.audio.start_playback()
        self.audio.wait_for_state_change().get()

        self.assertEvent("tags_changed", tags=mock.ANY)

    # Unlike the other events, having the state changed done is not
    # enough to ensure our event is called. So we setup a threading
    # event that we can wait for with a timeout while the track playback
    # completes.

    def test_stream_changed_event_on_paused(self):
        event = self.listener.wait("stream_changed").get()

        self.audio.prepare_change()
        self.audio.set_uri(self.uris[0])
        self.audio.pause_playback().get()
        self.audio.wait_for_state_change().get()

        if not event.wait(timeout=1.0):
            self.fail("Stream changed not reached within deadline")

        self.assertEvent("stream_changed", uri=self.uris[0])

    def test_reached_end_of_stream_event(self):
        event = self.listener.wait("reached_end_of_stream").get()

        self.audio.prepare_change()
        self.audio.set_uri(self.uris[0])
        self.audio.start_playback()
        self.audio.wait_for_state_change().get()

        self.possibly_trigger_fake_about_to_finish()
        if not event.wait(timeout=1.0):
            self.fail("End of stream not reached within deadline")

        assert not self.audio.get_current_tags().get()

    def test_gapless(self):
        uris = self.uris[1:]
        event = self.listener.wait("reached_end_of_stream").get()

        def callback():
            if uris:
                self.audio.set_uri(uris.pop()).get()

        self.audio.set_about_to_finish_callback(callback).get()

        self.audio.prepare_change()
        self.audio.set_uri(self.uris[0])
        self.audio.start_playback()

        self.possibly_trigger_fake_about_to_finish()
        self.audio.wait_for_state_change().get()

        self.possibly_trigger_fake_about_to_finish()
        self.audio.wait_for_state_change().get()
        if not event.wait(timeout=1.0):
            self.fail("EOS not received")

        # Check that both uris got played
        self.assertEvent("stream_changed", uri=self.uris[0])
        self.assertEvent("stream_changed", uri=self.uris[1])

        # Check that events counts check out.
        keys = [k for k, v in self.listener.get_events().get()]
        assert keys.count("stream_changed") == 2
        assert keys.count("position_changed") == 2
        assert keys.count("state_changed") == 1
        assert keys.count("reached_end_of_stream") == 1

        # TODO: test tag states within gaples

    # TODO: this does not belong in this testcase
    def test_current_tags_are_blank_to_begin_with(self):
        assert not self.audio.get_current_tags().get()

    def test_current_tags_blank_after_end_of_stream(self):
        event = self.listener.wait("reached_end_of_stream").get()

        self.audio.prepare_change()
        self.audio.set_uri(self.uris[0])
        self.audio.start_playback()

        self.possibly_trigger_fake_about_to_finish()
        self.audio.wait_for_state_change().get()

        if not event.wait(timeout=1.0):
            self.fail("EOS not received")

        assert not self.audio.get_current_tags().get()

    def test_current_tags_stored(self):
        event = self.listener.wait("reached_end_of_stream").get()
        tags = []

        def callback():
            tags.append(self.audio.get_current_tags().get())

        self.audio.set_about_to_finish_callback(callback).get()

        self.audio.prepare_change()
        self.audio.set_uri(self.uris[0])
        self.audio.start_playback()

        self.possibly_trigger_fake_about_to_finish()
        self.audio.wait_for_state_change().get()

        if not event.wait(timeout=1.0):
            self.fail("EOS not received")

        assert tags[0]

    # TODO: test that we reset when we expect between songs


class AudioDummyEventTest(DummyMixin, AudioEventTest):

    """Exercise the AudioEventTest against our mock audio classes."""


# TODO: move to mixer tests...
class MixerTest(BaseTest):
    @unittest.SkipTest
    def test_set_mute(self):
        for value in (True, False):
            assert self.audio.set_mute(value).get()
            assert self.audio.get_mute().get() == value

    @unittest.SkipTest
    def test_set_state_encapsulation(self):
        pass  # TODO

    @unittest.SkipTest
    def test_set_position(self):
        pass  # TODO

    @unittest.SkipTest
    def test_invalid_output_raises_error(self):
        pass  # TODO


class AudioStateTest(unittest.TestCase):
    def setUp(self):  # noqa: N802
        self.audio = audio.Audio(config=None, mixer=None)

    def test_state_starts_as_stopped(self):
        assert audio.PlaybackState.STOPPED == self.audio.state

    def test_state_does_not_change_when_in_gst_ready_state(self):
        self.audio._handler.on_playbin_state_changed(
            Gst.State.NULL, Gst.State.READY, Gst.State.VOID_PENDING
        )

        assert audio.PlaybackState.STOPPED == self.audio.state

    def test_state_changes_from_stopped_to_playing_on_play(self):
        self.audio._handler.on_playbin_state_changed(
            Gst.State.NULL, Gst.State.READY, Gst.State.PLAYING
        )
        self.audio._handler.on_playbin_state_changed(
            Gst.State.READY, Gst.State.PAUSED, Gst.State.PLAYING
        )
        self.audio._handler.on_playbin_state_changed(
            Gst.State.PAUSED, Gst.State.PLAYING, Gst.State.VOID_PENDING
        )

        assert audio.PlaybackState.PLAYING == self.audio.state

    def test_state_changes_from_playing_to_paused_on_pause(self):
        self.audio.state = audio.PlaybackState.PLAYING

        self.audio._handler.on_playbin_state_changed(
            Gst.State.PLAYING, Gst.State.PAUSED, Gst.State.VOID_PENDING
        )

        assert audio.PlaybackState.PAUSED == self.audio.state

    def test_state_changes_from_playing_to_stopped_on_stop(self):
        self.audio.state = audio.PlaybackState.PLAYING

        self.audio._handler.on_playbin_state_changed(
            Gst.State.PLAYING, Gst.State.PAUSED, Gst.State.NULL
        )
        self.audio._handler.on_playbin_state_changed(
            Gst.State.PAUSED, Gst.State.READY, Gst.State.NULL
        )
        # We never get the following call, so the logic must work without it
        # self.audio._handler.on_playbin_state_changed(
        #     Gst.State.READY, Gst.State.NULL, Gst.State.VOID_PENDING)

        assert audio.PlaybackState.STOPPED == self.audio.state


class AudioBufferingTest(unittest.TestCase):
    def setUp(self):  # noqa: N802
        self.audio = audio.Audio(config=None, mixer=None)
        self.audio._playbin = mock.Mock(spec=["set_state"])

    def test_pause_when_buffer_empty(self):
        playbin = self.audio._playbin
        self.audio.start_playback()
        playbin.set_state.assert_called_with(Gst.State.PLAYING)
        playbin.set_state.reset_mock()

        self.audio._handler.on_buffering(0)
        playbin.set_state.assert_called_with(Gst.State.PAUSED)
        assert self.audio._buffering

    def test_stay_paused_when_buffering_finished(self):
        playbin = self.audio._playbin
        self.audio.pause_playback()
        playbin.set_state.assert_called_with(Gst.State.PAUSED)
        playbin.set_state.reset_mock()

        self.audio._handler.on_buffering(100)
        assert playbin.set_state.call_count == 0
        assert not self.audio._buffering

    def test_change_to_paused_while_buffering(self):
        playbin = self.audio._playbin
        self.audio.start_playback()
        playbin.set_state.assert_called_with(Gst.State.PLAYING)
        playbin.set_state.reset_mock()

        self.audio._handler.on_buffering(0)
        playbin.set_state.assert_called_with(Gst.State.PAUSED)
        self.audio.pause_playback()
        playbin.set_state.reset_mock()

        self.audio._handler.on_buffering(100)
        assert playbin.set_state.call_count == 0
        assert not self.audio._buffering

    def test_change_to_stopped_while_buffering(self):
        playbin = self.audio._playbin
        self.audio.start_playback()
        playbin.set_state.assert_called_with(Gst.State.PLAYING)
        playbin.set_state.reset_mock()

        self.audio._handler.on_buffering(0)
        playbin.set_state.assert_called_with(Gst.State.PAUSED)
        playbin.set_state.reset_mock()

        self.audio.stop_playback()
        playbin.set_state.assert_called_with(Gst.State.NULL)
        assert not self.audio._buffering


class AudioLiveTest(unittest.TestCase):
    config = {
        "audio": {
            "buffer_time": None,
            "mixer": "fakemixer track_max_volume=65536",
            "mixer_track": None,
            "mixer_volume": None,
            "output": "testoutput",
            "visualizer": None,
        }
    }

    def setUp(self):  # noqa: N802
        config = {
            "audio": {
                "buffer_time": None,
                "mixer": "foomixer",
                "mixer_volume": None,
                "output": "testoutput",
                "visualizer": None,
            },
            "proxy": {"hostname": ""},
        }
        self.audio = audio.Audio(config=config, mixer=None)

    def test_not_live_mode(self):
        source = mock.MagicMock()

        # Avoid appsrc.configure()
        source.get_factory.get_name = mock.MagicMock(return_value="not_appsrc")

        source.props = mock.MagicMock(spec=[])
        self.audio._live_stream = False

        self.audio._on_source_setup("dummy", source)

        source.set_live.assert_not_called()

    def test_live_mode(self):
        source = mock.MagicMock()

        # Avoid appsrc.configure()
        source.get_factory.get_name = mock.MagicMock(return_value="not_appsrc")

        source.props.is_live = mock.MagicMock(return_value=True)
        self.audio._live_stream = True

        self.audio._on_source_setup("dummy", source)

        source.set_live.assert_called_with(True)


class DownloadBufferingTest(unittest.TestCase):
    def setUp(self):  # noqa: N802
        self.audio = audio.Audio(config=None, mixer=None)
        self.audio._playbin = mock.Mock(spec=["set_property"])

    def test_download_flag_is_passed_to_playbin_if_download_buffering_is_enabled(
        self,
    ):
        playbin = self.audio._playbin

        self.audio.set_uri("some:uri", False, True)

        playbin.set_property.assert_has_calls([mock.call("flags", 0x02 | 0x80)])

    def test_download_flag_is_not_passed_to_playbin_if_download_buffering_is_not_enabled(  # noqa: B950
        self,
    ):
        playbin = self.audio._playbin

        self.audio.set_uri("some:uri", False, False)

        playbin.set_property.assert_has_calls([mock.call("flags", 0x02)])



import json
import unittest

from mopidy.models import (
    Album,
    Artist,
    Image,
    ModelJSONEncoder,
    Playlist,
    Ref,
    SearchResult,
    TlTrack,
    Track,
    model_json_decoder,
)


class InheritanceTest(unittest.TestCase):
    def test_weakref_and_slots_play_nice_in_subclass(self):
        # Check that the following does not happen:
        # TypeError: Error when calling the metaclass bases
        #   __weakref__ slot disallowed: either we already got one...

        class Foo(Track):
            pass

    def test_sub_class_can_have_its_own_slots(self):
        # Needed for things like SpotifyTrack in mopidy-spotify 1.x

        class Foo(Track):
            __slots__ = ("_foo",)

        f = Foo()
        f._foo = 123

    def test_sub_class_can_be_initialized(self):
        # Fails with following error if fields are not handled across classes.
        #   TypeError: __init__() got an unexpected keyword argument "type"
        # Essentially this is testing that sub-classes take parent _fields into
        # account.

        class Foo(Ref):
            pass

        Foo.directory()


class CachingTest(unittest.TestCase):
    def test_same_instance(self):
        assert Track() is Track()

    def test_same_instance_with_values(self):
        assert Track(uri="test") is Track(uri="test")

    def test_different_instance_with_different_values(self):
        assert Track(uri="test1") is not Track(uri="test2")

    def test_different_instance_with_replace(self):
        t = Track(uri="test1")
        assert t is not t.replace(uri="test2")


class GenericReplaceTest(unittest.TestCase):
    def compare(self, orig, other):
        assert orig == other
        assert id(orig) == id(other)

    def test_replace_track(self):
        track = Track()
        self.compare(track, track.replace())

    def test_replace_artist(self):
        artist = Artist()
        self.compare(artist, artist.replace())

    def test_replace_album(self):
        album = Album()
        self.compare(album, album.replace())

    def test_replace_playlist(self):
        playlist = Playlist()
        self.compare(playlist, playlist.replace())

    def test_replace_track_with_basic_values(self):
        track = Track(name="foo", uri="bar")
        other = track.replace(name="baz")
        assert "baz" == other.name
        assert "bar" == other.uri

    def test_replace_track_with_missing_values(self):
        track = Track(uri="bar")
        other = track.replace(name="baz")
        assert "baz" == other.name
        assert "bar" == other.uri

    def test_replace_track_with_private_internal_value(self):
        artist1 = Artist(name="foo")
        artist2 = Artist(name="bar")
        track = Track(artists=[artist1])
        other = track.replace(artists=[artist2])
        assert artist2 in other.artists

    def test_replace_track_with_invalid_key(self):
        with self.assertRaises(TypeError):
            Track().replace(invalid_key=True)

    def test_replace_track_to_remove(self):
        track = Track(name="foo").replace(name=None)
        assert not hasattr(track, "_name")


class RefTest(unittest.TestCase):
    def test_uri(self):
        uri = "an_uri"
        ref = Ref(uri=uri)
        assert ref.uri == uri
        with self.assertRaises(AttributeError):
            ref.uri = None

    def test_name(self):
        name = "a name"
        ref = Ref(name=name)
        assert ref.name == name
        with self.assertRaises(AttributeError):
            ref.name = None

    # TODO: add these for the more of the models?
    def test_del_name(self):
        ref = Ref(name="foo")
        with self.assertRaises(AttributeError):
            del ref.name

    def test_invalid_kwarg(self):
        with self.assertRaises(TypeError):
            Ref(foo="baz")

    def test_repr_without_results(self):
        assert "Ref(name='foo', type='artist', uri='uri')" == repr(
            Ref(uri="uri", name="foo", type="artist")
        )

    def test_serialize_without_results(self):
        self.assertDictEqual(
            {"__model__": "Ref", "uri": "uri"}, Ref(uri="uri").serialize()
        )

    def test_to_json_and_back(self):
        ref1 = Ref(uri="uri")
        serialized = json.dumps(ref1, cls=ModelJSONEncoder)
        ref2 = json.loads(serialized, object_hook=model_json_decoder)
        assert ref1 == ref2

    def test_type_constants(self):
        assert Ref.ALBUM == "album"
        assert Ref.ARTIST == "artist"
        assert Ref.DIRECTORY == "directory"
        assert Ref.PLAYLIST == "playlist"
        assert Ref.TRACK == "track"

    def test_album_constructor(self):
        ref = Ref.album(uri="foo", name="bar")
        assert ref.uri == "foo"
        assert ref.name == "bar"
        assert ref.type == Ref.ALBUM

    def test_artist_constructor(self):
        ref = Ref.artist(uri="foo", name="bar")
        assert ref.uri == "foo"
        assert ref.name == "bar"
        assert ref.type == Ref.ARTIST

    def test_directory_constructor(self):
        ref = Ref.directory(uri="foo", name="bar")
        assert ref.uri == "foo"
        assert ref.name == "bar"
        assert ref.type == Ref.DIRECTORY

    def test_playlist_constructor(self):
        ref = Ref.playlist(uri="foo", name="bar")
        assert ref.uri == "foo"
        assert ref.name == "bar"
        assert ref.type == Ref.PLAYLIST

    def test_track_constructor(self):
        ref = Ref.track(uri="foo", name="bar")
        assert ref.uri == "foo"
        assert ref.name == "bar"
        assert ref.type == Ref.TRACK


class ImageTest(unittest.TestCase):
    def test_uri(self):
        uri = "an_uri"
        image = Image(uri=uri)
        assert image.uri == uri
        with self.assertRaises(AttributeError):
            image.uri = None

    def test_width(self):
        image = Image(width=100)
        assert image.width == 100
        with self.assertRaises(AttributeError):
            image.width = None

    def test_height(self):
        image = Image(height=100)
        assert image.height == 100
        with self.assertRaises(AttributeError):
            image.height = None

    def test_invalid_kwarg(self):
        with self.assertRaises(TypeError):
            Image(foo="baz")


class ArtistTest(unittest.TestCase):
    def test_uri(self):
        uri = "an_uri"
        artist = Artist(uri=uri)
        assert artist.uri == uri
        with self.assertRaises(AttributeError):
            artist.uri = None

    def test_name(self):
        name = "a name"
        artist = Artist(name=name)
        assert artist.name == name
        with self.assertRaises(AttributeError):
            artist.name = None

    def test_musicbrainz_id(self):
        mb_id = "mb-id"
        artist = Artist(musicbrainz_id=mb_id)
        assert artist.musicbrainz_id == mb_id
        with self.assertRaises(AttributeError):
            artist.musicbrainz_id = None

    def test_invalid_kwarg(self):
        with self.assertRaises(TypeError):
            Artist(foo="baz")

    def test_invalid_kwarg_with_name_matching_method(self):
        with self.assertRaises(TypeError):
            Artist(replace="baz")

        with self.assertRaises(TypeError):
            Artist(serialize="baz")

    def test_repr(self):
        assert "Artist(name='name', uri='uri')" == repr(
            Artist(uri="uri", name="name")
        )

    def test_serialize(self):
        self.assertDictEqual(
            {"__model__": "Artist", "uri": "uri", "name": "name"},
            Artist(uri="uri", name="name").serialize(),
        )

    def test_serialize_falsy_values(self):
        self.assertDictEqual(
            {"__model__": "Artist", "uri": "", "name": ""},
            Artist(uri="", name="").serialize(),
        )

    def test_to_json_and_back(self):
        artist1 = Artist(uri="uri", name="name")
        serialized = json.dumps(artist1, cls=ModelJSONEncoder)
        artist2 = json.loads(serialized, object_hook=model_json_decoder)
        assert artist1 == artist2

    def test_to_json_and_back_with_unknown_field(self):
        artist = Artist(uri="uri", name="name").serialize()
        artist["foo"] = "foo"
        serialized = json.dumps(artist)
        with self.assertRaises(TypeError):
            json.loads(serialized, object_hook=model_json_decoder)

    def test_to_json_and_back_with_field_matching_method(self):
        artist = Artist(uri="uri", name="name").serialize()
        artist["copy"] = "foo"
        serialized = json.dumps(artist)
        with self.assertRaises(TypeError):
            json.loads(serialized, object_hook=model_json_decoder)

    def test_to_json_and_back_with_field_matching_internal_field(self):
        artist = Artist(uri="uri", name="name").serialize()
        artist["__mro__"] = "foo"
        serialized = json.dumps(artist)
        with self.assertRaises(TypeError):
            json.loads(serialized, object_hook=model_json_decoder)

    def test_eq_name(self):
        artist1 = Artist(name="name")
        artist2 = Artist(name="name")
        assert artist1 == artist2
        assert hash(artist1) == hash(artist2)

    def test_eq_uri(self):
        artist1 = Artist(uri="uri")
        artist2 = Artist(uri="uri")
        assert artist1 == artist2
        assert hash(artist1) == hash(artist2)

    def test_eq_musibrainz_id(self):
        artist1 = Artist(musicbrainz_id="id")
        artist2 = Artist(musicbrainz_id="id")
        assert artist1 == artist2
        assert hash(artist1) == hash(artist2)

    def test_eq(self):
        artist1 = Artist(uri="uri", name="name", musicbrainz_id="id")
        artist2 = Artist(uri="uri", name="name", musicbrainz_id="id")
        assert artist1 == artist2
        assert hash(artist1) == hash(artist2)

    def test_eq_none(self):
        assert Artist() is not None

    def test_eq_other(self):
        assert Artist() != "other"

    def test_ne_name(self):
        artist1 = Artist(name="name1")
        artist2 = Artist(name="name2")
        assert artist1 != artist2
        assert hash(artist1) != hash(artist2)

    def test_ne_uri(self):
        artist1 = Artist(uri="uri1")
        artist2 = Artist(uri="uri2")
        assert artist1 != artist2
        assert hash(artist1) != hash(artist2)

    def test_ne_musicbrainz_id(self):
        artist1 = Artist(musicbrainz_id="id1")
        artist2 = Artist(musicbrainz_id="id2")
        assert artist1 != artist2
        assert hash(artist1) != hash(artist2)

    def test_ne(self):
        artist1 = Artist(uri="uri1", name="name1", musicbrainz_id="id1")
        artist2 = Artist(uri="uri2", name="name2", musicbrainz_id="id2")
        assert artist1 != artist2
        assert hash(artist1) != hash(artist2)


class AlbumTest(unittest.TestCase):
    def test_uri(self):
        uri = "an_uri"
        album = Album(uri=uri)
        assert album.uri == uri
        with self.assertRaises(AttributeError):
            album.uri = None

    def test_name(self):
        name = "a name"
        album = Album(name=name)
        assert album.name == name
        with self.assertRaises(AttributeError):
            album.name = None

    def test_artists(self):
        artist = Artist()
        album = Album(artists=[artist])
        assert artist in album.artists
        with self.assertRaises(AttributeError):
            album.artists = None

    def test_artists_none(self):
        assert set() == Album(artists=None).artists

    def test_num_tracks(self):
        num_tracks = 11
        album = Album(num_tracks=num_tracks)
        assert album.num_tracks == num_tracks
        with self.assertRaises(AttributeError):
            album.num_tracks = None

    def test_num_discs(self):
        num_discs = 2
        album = Album(num_discs=num_discs)
        assert album.num_discs == num_discs
        with self.assertRaises(AttributeError):
            album.num_discs = None

    def test_date(self):
        date = "1977-01-01"
        album = Album(date=date)
        assert album.date == date
        with self.assertRaises(AttributeError):
            album.date = None

    def test_musicbrainz_id(self):
        mb_id = "mb-id"
        album = Album(musicbrainz_id=mb_id)
        assert album.musicbrainz_id == mb_id
        with self.assertRaises(AttributeError):
            album.musicbrainz_id = None

    def test_invalid_kwarg(self):
        with self.assertRaises(TypeError):
            Album(foo="baz")

    def test_repr_without_artists(self):
        assert "Album(name='name', uri='uri')" == repr(
            Album(uri="uri", name="name")
        )

    def test_repr_with_artists(self):
        assert (
            "Album(artists=[Artist(name='foo')], name='name', uri='uri')"
            == repr(Album(uri="uri", name="name", artists=[Artist(name="foo")]))
        )

    def test_serialize_without_artists(self):
        self.assertDictEqual(
            {"__model__": "Album", "uri": "uri", "name": "name"},
            Album(uri="uri", name="name").serialize(),
        )

    def test_serialize_with_artists(self):
        artist = Artist(name="foo")
        self.assertDictEqual(
            {
                "__model__": "Album",
                "uri": "uri",
                "name": "name",
                "artists": [artist.serialize()],
            },
            Album(uri="uri", name="name", artists=[artist]).serialize(),
        )

    def test_to_json_and_back(self):
        album1 = Album(uri="uri", name="name", artists=[Artist(name="foo")])
        serialized = json.dumps(album1, cls=ModelJSONEncoder)
        album2 = json.loads(serialized, object_hook=model_json_decoder)
        assert album1 == album2

    def test_eq_name(self):
        album1 = Album(name="name")
        album2 = Album(name="name")
        assert album1 == album2
        assert hash(album1) == hash(album2)

    def test_eq_uri(self):
        album1 = Album(uri="uri")
        album2 = Album(uri="uri")
        assert album1 == album2
        assert hash(album1) == hash(album2)

    def test_eq_artists(self):
        artists = [Artist()]
        album1 = Album(artists=artists)
        album2 = Album(artists=artists)
        assert album1 == album2
        assert hash(album1) == hash(album2)

    def test_eq_artists_order(self):
        artist1 = Artist(name="name1")
        artist2 = Artist(name="name2")
        album1 = Album(artists=[artist1, artist2])
        album2 = Album(artists=[artist2, artist1])
        assert album1 == album2
        assert hash(album1) == hash(album2)

    def test_eq_num_tracks(self):
        album1 = Album(num_tracks=2)
        album2 = Album(num_tracks=2)
        assert album1 == album2
        assert hash(album1) == hash(album2)

    def test_eq_date(self):
        date = "1977-01-01"
        album1 = Album(date=date)
        album2 = Album(date=date)
        assert album1 == album2
        assert hash(album1) == hash(album2)

    def test_eq_musibrainz_id(self):
        album1 = Album(musicbrainz_id="id")
        album2 = Album(musicbrainz_id="id")
        assert album1 == album2
        assert hash(album1) == hash(album2)

    def test_eq(self):
        artists = [Artist()]
        album1 = Album(
            name="name",
            uri="uri",
            artists=artists,
            num_tracks=2,
            musicbrainz_id="id",
        )
        album2 = Album(
            name="name",
            uri="uri",
            artists=artists,
            num_tracks=2,
            musicbrainz_id="id",
        )
        assert album1 == album2
        assert hash(album1) == hash(album2)

    def test_eq_none(self):
        assert Album() is not None

    def test_eq_other(self):
        assert Album() != "other"

    def test_ne_name(self):
        album1 = Album(name="name1")
        album2 = Album(name="name2")
        assert album1 != album2
        assert hash(album1) != hash(album2)

    def test_ne_uri(self):
        album1 = Album(uri="uri1")
        album2 = Album(uri="uri2")
        assert album1 != album2
        assert hash(album1) != hash(album2)

    def test_ne_artists(self):
        album1 = Album(artists=[Artist(name="name1")])
        album2 = Album(artists=[Artist(name="name2")])
        assert album1 != album2
        assert hash(album1) != hash(album2)

    def test_ne_num_tracks(self):
        album1 = Album(num_tracks=1)
        album2 = Album(num_tracks=2)
        assert album1 != album2
        assert hash(album1) != hash(album2)

    def test_ne_date(self):
        album1 = Album(date="1977-01-01")
        album2 = Album(date="1977-01-02")
        assert album1 != album2
        assert hash(album1) != hash(album2)

    def test_ne_musicbrainz_id(self):
        album1 = Album(musicbrainz_id="id1")
        album2 = Album(musicbrainz_id="id2")
        assert album1 != album2
        assert hash(album1) != hash(album2)

    def test_ne(self):
        album1 = Album(
            name="name1",
            uri="uri1",
            artists=[Artist(name="name1")],
            num_tracks=1,
            musicbrainz_id="id1",
        )
        album2 = Album(
            name="name2",
            uri="uri2",
            artists=[Artist(name="name2")],
            num_tracks=2,
            musicbrainz_id="id2",
        )
        assert album1 != album2
        assert hash(album1) != hash(album2)


class TrackTest(unittest.TestCase):
    def test_uri(self):
        uri = "an_uri"
        track = Track(uri=uri)
        assert track.uri == uri
        with self.assertRaises(AttributeError):
            track.uri = None

    def test_name(self):
        name = "a name"
        track = Track(name=name)
        assert track.name == name
        with self.assertRaises(AttributeError):
            track.name = None

    def test_artists(self):
        artists = [Artist(name="name1"), Artist(name="name2")]
        track = Track(artists=artists)
        assert set(track.artists) == set(artists)
        with self.assertRaises(AttributeError):
            track.artists = None

    def test_artists_none(self):
        assert set() == Track(artists=None).artists

    def test_composers(self):
        artists = [Artist(name="name1"), Artist(name="name2")]
        track = Track(composers=artists)
        assert set(track.composers) == set(artists)
        with self.assertRaises(AttributeError):
            track.composers = None

    def test_composers_none(self):
        assert set() == Track(composers=None).composers

    def test_performers(self):
        artists = [Artist(name="name1"), Artist(name="name2")]
        track = Track(performers=artists)
        assert set(track.performers) == set(artists)
        with self.assertRaises(AttributeError):
            track.performers = None

    def test_performers_none(self):
        assert set() == Track(performers=None).performers

    def test_album(self):
        album = Album()
        track = Track(album=album)
        assert track.album == album
        with self.assertRaises(AttributeError):
            track.album = None

    def test_track_no(self):
        track_no = 7
        track = Track(track_no=track_no)
        assert track.track_no == track_no
        with self.assertRaises(AttributeError):
            track.track_no = None

    def test_disc_no(self):
        disc_no = 2
        track = Track(disc_no=disc_no)
        assert track.disc_no == disc_no
        with self.assertRaises(AttributeError):
            track.disc_no = None

    def test_date(self):
        date = "1977-01-01"
        track = Track(date=date)
        assert track.date == date
        with self.assertRaises(AttributeError):
            track.date = None

    def test_length(self):
        length = 137000
        track = Track(length=length)
        assert track.length == length
        with self.assertRaises(AttributeError):
            track.length = None

    def test_bitrate(self):
        bitrate = 160
        track = Track(bitrate=bitrate)
        assert track.bitrate == bitrate
        with self.assertRaises(AttributeError):
            track.bitrate = None

    def test_musicbrainz_id(self):
        mb_id = "mb-id"
        track = Track(musicbrainz_id=mb_id)
        assert track.musicbrainz_id == mb_id
        with self.assertRaises(AttributeError):
            track.musicbrainz_id = None

    def test_invalid_kwarg(self):
        with self.assertRaises(TypeError):
            Track(foo="baz")

    def test_repr_without_artists(self):
        assert "Track(name='name', uri='uri')" == repr(
            Track(uri="uri", name="name")
        )

    def test_repr_with_artists(self):
        assert (
            "Track(artists=[Artist(name='foo')], name='name', uri='uri')"
            == repr(Track(uri="uri", name="name", artists=[Artist(name="foo")]))
        )

    def test_serialize_without_artists(self):
        self.assertDictEqual(
            {"__model__": "Track", "uri": "uri", "name": "name"},
            Track(uri="uri", name="name").serialize(),
        )

    def test_serialize_with_artists(self):
        artist = Artist(name="foo")
        self.assertDictEqual(
            {
                "__model__": "Track",
                "uri": "uri",
                "name": "name",
                "artists": [artist.serialize()],
            },
            Track(uri="uri", name="name", artists=[artist]).serialize(),
        )

    def test_serialize_with_album(self):
        album = Album(name="foo")
        self.assertDictEqual(
            {
                "__model__": "Track",
                "uri": "uri",
                "name": "name",
                "album": album.serialize(),
            },
            Track(uri="uri", name="name", album=album).serialize(),
        )

    def test_to_json_and_back(self):
        track1 = Track(
            uri="uri",
            name="name",
            album=Album(name="foo"),
            artists=[Artist(name="foo")],
        )
        serialized = json.dumps(track1, cls=ModelJSONEncoder)
        track2 = json.loads(serialized, object_hook=model_json_decoder)
        assert track1 == track2

    def test_eq_uri(self):
        track1 = Track(uri="uri1")
        track2 = Track(uri="uri1")
        assert track1 == track2
        assert hash(track1) == hash(track2)

    def test_eq_name(self):
        track1 = Track(name="name1")
        track2 = Track(name="name1")
        assert track1 == track2
        assert hash(track1) == hash(track2)

    def test_eq_artists(self):
        artists = [Artist()]
        track1 = Track(artists=artists)
        track2 = Track(artists=artists)
        assert track1 == track2
        assert hash(track1) == hash(track2)

    def test_eq_artists_order(self):
        artist1 = Artist(name="name1")
        artist2 = Artist(name="name2")
        track1 = Track(artists=[artist1, artist2])
        track2 = Track(artists=[artist2, artist1])
        assert track1 == track2
        assert hash(track1) == hash(track2)

    def test_eq_album(self):
        album = Album()
        track1 = Track(album=album)
        track2 = Track(album=album)
        assert track1 == track2
        assert hash(track1) == hash(track2)

    def test_eq_track_no(self):
        track1 = Track(track_no=1)
        track2 = Track(track_no=1)
        assert track1 == track2
        assert hash(track1) == hash(track2)

    def test_eq_date(self):
        date = "1977-01-01"
        track1 = Track(date=date)
        track2 = Track(date=date)
        assert track1 == track2
        assert hash(track1) == hash(track2)

    def test_eq_length(self):
        track1 = Track(length=100)
        track2 = Track(length=100)
        assert track1 == track2
        assert hash(track1) == hash(track2)

    def test_eq_bitrate(self):
        track1 = Track(bitrate=100)
        track2 = Track(bitrate=100)
        assert track1 == track2
        assert hash(track1) == hash(track2)

    def test_eq_musibrainz_id(self):
        track1 = Track(musicbrainz_id="id")
        track2 = Track(musicbrainz_id="id")
        assert track1 == track2
        assert hash(track1) == hash(track2)

    def test_eq(self):
        date = "1977-01-01"
        artists = [Artist()]
        album = Album()
        track1 = Track(
            uri="uri",
            name="name",
            artists=artists,
            album=album,
            track_no=1,
            date=date,
            length=100,
            bitrate=100,
            musicbrainz_id="id",
        )
        track2 = Track(
            uri="uri",
            name="name",
            artists=artists,
            album=album,
            track_no=1,
            date=date,
            length=100,
            bitrate=100,
            musicbrainz_id="id",
        )
        assert track1 == track2
        assert hash(track1) == hash(track2)

    def test_eq_none(self):
        assert Track() is not None

    def test_eq_other(self):
        assert Track() != "other"

    def test_ne_uri(self):
        track1 = Track(uri="uri1")
        track2 = Track(uri="uri2")
        assert track1 != track2
        assert hash(track1) != hash(track2)

    def test_ne_name(self):
        track1 = Track(name="name1")
        track2 = Track(name="name2")
        assert track1 != track2
        assert hash(track1) != hash(track2)

    def test_ne_artists(self):
        track1 = Track(artists=[Artist(name="name1")])
        track2 = Track(artists=[Artist(name="name2")])
        assert track1 != track2
        assert hash(track1) != hash(track2)

    def test_ne_album(self):
        track1 = Track(album=Album(name="name1"))
        track2 = Track(album=Album(name="name2"))
        assert track1 != track2
        assert hash(track1) != hash(track2)

    def test_ne_track_no(self):
        track1 = Track(track_no=1)
        track2 = Track(track_no=2)
        assert track1 != track2
        assert hash(track1) != hash(track2)

    def test_ne_date(self):
        track1 = Track(date="1977-01-01")
        track2 = Track(date="1977-01-02")
        assert track1 != track2
        assert hash(track1) != hash(track2)

    def test_ne_length(self):
        track1 = Track(length=100)
        track2 = Track(length=200)
        assert track1 != track2
        assert hash(track1) != hash(track2)

    def test_ne_bitrate(self):
        track1 = Track(bitrate=100)
        track2 = Track(bitrate=200)
        assert track1 != track2
        assert hash(track1) != hash(track2)

    def test_ne_musicbrainz_id(self):
        track1 = Track(musicbrainz_id="id1")
        track2 = Track(musicbrainz_id="id2")
        assert track1 != track2
        assert hash(track1) != hash(track2)

    def test_ne(self):
        track1 = Track(
            uri="uri1",
            name="name1",
            artists=[Artist(name="name1")],
            album=Album(name="name1"),
            track_no=1,
            date="1977-01-01",
            length=100,
            bitrate=100,
            musicbrainz_id="id1",
        )
        track2 = Track(
            uri="uri2",
            name="name2",
            artists=[Artist(name="name2")],
            album=Album(name="name2"),
            track_no=2,
            date="1977-01-02",
            length=200,
            bitrate=200,
            musicbrainz_id="id2",
        )
        assert track1 != track2
        assert hash(track1) != hash(track2)

    def test_ignores_values_with_default_value_none(self):
        track1 = Track(name="name1")
        track2 = Track(name="name1", album=None)
        assert track1 == track2
        assert hash(track1) == hash(track2)

    def test_replace_can_reset_to_default_value(self):
        track1 = Track(name="name1")
        track2 = Track(name="name1", album=Album()).replace(album=None)
        assert track1 == track2
        assert hash(track1) == hash(track2)


class TlTrackTest(unittest.TestCase):
    def test_tlid(self):
        tlid = 123
        tl_track = TlTrack(tlid=tlid)
        assert tl_track.tlid == tlid
        with self.assertRaises(AttributeError):
            tl_track.tlid = None

    def test_track(self):
        track = Track()
        tl_track = TlTrack(track=track)
        assert tl_track.track == track
        with self.assertRaises(AttributeError):
            tl_track.track = None

    def test_invalid_kwarg(self):
        with self.assertRaises(TypeError):
            TlTrack(foo="baz")

    def test_positional_args(self):
        tlid = 123
        track = Track()
        tl_track = TlTrack(tlid, track)
        assert tl_track.tlid == tlid
        assert tl_track.track == track

    def test_iteration(self):
        tlid = 123
        track = Track()
        tl_track = TlTrack(tlid, track)
        (tlid2, track2) = tl_track
        assert tlid2 == tlid
        assert track2 == track

    def test_repr(self):
        assert "TlTrack(tlid=123, track=Track(uri='uri'))" == repr(
            TlTrack(tlid=123, track=Track(uri="uri"))
        )

    def test_serialize(self):
        track = Track(uri="uri", name="name")
        self.assertDictEqual(
            {"__model__": "TlTrack", "tlid": 123, "track": track.serialize()},
            TlTrack(tlid=123, track=track).serialize(),
        )

    def test_to_json_and_back(self):
        tl_track1 = TlTrack(tlid=123, track=Track(uri="uri", name="name"))
        serialized = json.dumps(tl_track1, cls=ModelJSONEncoder)
        tl_track2 = json.loads(serialized, object_hook=model_json_decoder)
        assert tl_track1 == tl_track2

    def test_eq(self):
        tlid = 123
        track = Track()
        tl_track1 = TlTrack(tlid=tlid, track=track)
        tl_track2 = TlTrack(tlid=tlid, track=track)
        assert tl_track1 == tl_track2
        assert hash(tl_track1) == hash(tl_track2)

    def test_eq_none(self):
        assert TlTrack() is not None

    def test_eq_other(self):
        assert TlTrack() != "other"

    def test_ne_tlid(self):
        tl_track1 = TlTrack(tlid=123)
        tl_track2 = TlTrack(tlid=321)
        assert tl_track1 != tl_track2
        assert hash(tl_track1) != hash(tl_track2)

    def test_ne_track(self):
        tl_track1 = TlTrack(track=Track(uri="a"))
        tl_track2 = TlTrack(track=Track(uri="b"))
        assert tl_track1 != tl_track2
        assert hash(tl_track1) != hash(tl_track2)


class PlaylistTest(unittest.TestCase):
    def test_uri(self):
        uri = "an_uri"
        playlist = Playlist(uri=uri)
        assert playlist.uri == uri
        with self.assertRaises(AttributeError):
            playlist.uri = None

    def test_name(self):
        name = "a name"
        playlist = Playlist(name=name)
        assert playlist.name == name
        with self.assertRaises(AttributeError):
            playlist.name = None

    def test_tracks(self):
        tracks = [Track(), Track(), Track()]
        playlist = Playlist(tracks=tracks)
        assert list(playlist.tracks) == tracks
        with self.assertRaises(AttributeError):
            playlist.tracks = None

    def test_length(self):
        tracks = [Track(), Track(), Track()]
        playlist = Playlist(tracks=tracks)
        assert playlist.length == 3

    def test_last_modified(self):
        last_modified = 1390942873000
        playlist = Playlist(last_modified=last_modified)
        assert playlist.last_modified == last_modified
        with self.assertRaises(AttributeError):
            playlist.last_modified = None

    def test_with_new_uri(self):
        tracks = [Track()]
        last_modified = 1390942873000
        playlist = Playlist(
            uri="an uri",
            name="a name",
            tracks=tracks,
            last_modified=last_modified,
        )
        new_playlist = playlist.replace(uri="another uri")
        assert new_playlist.uri == "another uri"
        assert new_playlist.name == "a name"
        assert list(new_playlist.tracks) == tracks
        assert new_playlist.last_modified == last_modified

    def test_with_new_name(self):
        tracks = [Track()]
        last_modified = 1390942873000
        playlist = Playlist(
            uri="an uri",
            name="a name",
            tracks=tracks,
            last_modified=last_modified,
        )
        new_playlist = playlist.replace(name="another name")
        assert new_playlist.uri == "an uri"
        assert new_playlist.name == "another name"
        assert list(new_playlist.tracks) == tracks
        assert new_playlist.last_modified == last_modified

    def test_with_new_tracks(self):
        tracks = [Track()]
        last_modified = 1390942873000
        playlist = Playlist(
            uri="an uri",
            name="a name",
            tracks=tracks,
            last_modified=last_modified,
        )
        new_tracks = [Track(), Track()]
        new_playlist = playlist.replace(tracks=new_tracks)
        assert new_playlist.uri == "an uri"
        assert new_playlist.name == "a name"
        assert list(new_playlist.tracks) == new_tracks
        assert new_playlist.last_modified == last_modified

    def test_with_new_last_modified(self):
        tracks = [Track()]
        last_modified = 1390942873000
        new_last_modified = last_modified + 1000
        playlist = Playlist(
            uri="an uri",
            name="a name",
            tracks=tracks,
            last_modified=last_modified,
        )
        new_playlist = playlist.replace(last_modified=new_last_modified)
        assert new_playlist.uri == "an uri"
        assert new_playlist.name == "a name"
        assert list(new_playlist.tracks) == tracks
        assert new_playlist.last_modified == new_last_modified

    def test_invalid_kwarg(self):
        with self.assertRaises(TypeError):
            Playlist(foo="baz")

    def test_repr_without_tracks(self):
        assert "Playlist(name='name', uri='uri')" == repr(
            Playlist(uri="uri", name="name")
        )

    def test_repr_with_tracks(self):
        assert (
            "Playlist(name='name', tracks=[Track(name='foo')], uri='uri')"
            == repr(
                Playlist(uri="uri", name="name", tracks=[Track(name="foo")])
            )
        )

    def test_serialize_without_tracks(self):
        self.assertDictEqual(
            {"__model__": "Playlist", "uri": "uri", "name": "name"},
            Playlist(uri="uri", name="name").serialize(),
        )

    def test_serialize_with_tracks(self):
        track = Track(name="foo")
        self.assertDictEqual(
            {
                "__model__": "Playlist",
                "uri": "uri",
                "name": "name",
                "tracks": [track.serialize()],
            },
            Playlist(uri="uri", name="name", tracks=[track]).serialize(),
        )

    def test_to_json_and_back(self):
        playlist1 = Playlist(uri="uri", name="name")
        serialized = json.dumps(playlist1, cls=ModelJSONEncoder)
        playlist2 = json.loads(serialized, object_hook=model_json_decoder)
        assert playlist1 == playlist2

    def test_eq_name(self):
        playlist1 = Playlist(name="name")
        playlist2 = Playlist(name="name")
        assert playlist1 == playlist2
        assert hash(playlist1) == hash(playlist2)

    def test_eq_uri(self):
        playlist1 = Playlist(uri="uri")
        playlist2 = Playlist(uri="uri")
        assert playlist1 == playlist2
        assert hash(playlist1) == hash(playlist2)

    def test_eq_tracks(self):
        tracks = [Track()]
        playlist1 = Playlist(tracks=tracks)
        playlist2 = Playlist(tracks=tracks)
        assert playlist1 == playlist2
        assert hash(playlist1) == hash(playlist2)

    def test_eq_last_modified(self):
        playlist1 = Playlist(last_modified=1)
        playlist2 = Playlist(last_modified=1)
        assert playlist1 == playlist2
        assert hash(playlist1) == hash(playlist2)

    def test_eq(self):
        tracks = [Track()]
        playlist1 = Playlist(
            uri="uri", name="name", tracks=tracks, last_modified=1
        )
        playlist2 = Playlist(
            uri="uri", name="name", tracks=tracks, last_modified=1
        )
        assert playlist1 == playlist2
        assert hash(playlist1) == hash(playlist2)

    def test_eq_none(self):
        assert Playlist() is not None

    def test_eq_other(self):
        assert Playlist() != "other"

    def test_ne_name(self):
        playlist1 = Playlist(name="name1")
        playlist2 = Playlist(name="name2")
        assert playlist1 != playlist2
        assert hash(playlist1) != hash(playlist2)

    def test_ne_uri(self):
        playlist1 = Playlist(uri="uri1")
        playlist2 = Playlist(uri="uri2")
        assert playlist1 != playlist2
        assert hash(playlist1) != hash(playlist2)

    def test_ne_tracks(self):
        playlist1 = Playlist(tracks=[Track(uri="uri1")])
        playlist2 = Playlist(tracks=[Track(uri="uri2")])
        assert playlist1 != playlist2
        assert hash(playlist1) != hash(playlist2)

    def test_ne_last_modified(self):
        playlist1 = Playlist(last_modified=1)
        playlist2 = Playlist(last_modified=2)
        assert playlist1 != playlist2
        assert hash(playlist1) != hash(playlist2)

    def test_ne(self):
        playlist1 = Playlist(
            uri="uri1",
            name="name1",
            tracks=[Track(uri="uri1")],
            last_modified=1,
        )
        playlist2 = Playlist(
            uri="uri2",
            name="name2",
            tracks=[Track(uri="uri2")],
            last_modified=2,
        )
        assert playlist1 != playlist2
        assert hash(playlist1) != hash(playlist2)


class SearchResultTest(unittest.TestCase):
    def test_uri(self):
        uri = "an_uri"
        result = SearchResult(uri=uri)
        assert result.uri == uri
        with self.assertRaises(AttributeError):
            result.uri = None

    def test_tracks(self):
        tracks = [Track(), Track(), Track()]
        result = SearchResult(tracks=tracks)
        assert list(result.tracks) == tracks
        with self.assertRaises(AttributeError):
            result.tracks = None

    def test_artists(self):
        artists = [Artist(), Artist(), Artist()]
        result = SearchResult(artists=artists)
        assert list(result.artists) == artists
        with self.assertRaises(AttributeError):
            result.artists = None

    def test_albums(self):
        albums = [Album(), Album(), Album()]
        result = SearchResult(albums=albums)
        assert list(result.albums) == albums
        with self.assertRaises(AttributeError):
            result.albums = None

    def test_invalid_kwarg(self):
        with self.assertRaises(TypeError):
            SearchResult(foo="baz")

    def test_repr_without_results(self):
        assert "SearchResult(uri='uri')" == repr(SearchResult(uri="uri"))

    def test_serialize_without_results(self):
        self.assertDictEqual(
            {"__model__": "SearchResult", "uri": "uri"},
            SearchResult(uri="uri").serialize(),
        )



import re

import pytest

from mopidy import httpclient


@pytest.mark.parametrize(
    "config,expected",
    [
        ({}, None),
        ({"hostname": ""}, None),
        ({"hostname": "proxy.lan"}, "http://proxy.lan:80"),
        ({"scheme": None, "hostname": "proxy.lan"}, "http://proxy.lan:80"),
        ({"scheme": "https", "hostname": "proxy.lan"}, "https://proxy.lan:80"),
        ({"username": "user", "hostname": "proxy.lan"}, "http://proxy.lan:80"),
        ({"password": "pass", "hostname": "proxy.lan"}, "http://proxy.lan:80"),
        ({"hostname": "proxy.lan", "port": 8080}, "http://proxy.lan:8080"),
        ({"hostname": "proxy.lan", "port": -1}, "http://proxy.lan:80"),
        ({"hostname": "proxy.lan", "port": None}, "http://proxy.lan:80"),
        ({"hostname": "proxy.lan", "port": ""}, "http://proxy.lan:80"),
        (
            {"username": "user", "password": "pass", "hostname": "proxy.lan"},
            "http://user:pass@proxy.lan:80",
        ),
    ],
)
def test_format_proxy(config, expected):
    assert httpclient.format_proxy(config) == expected


def test_format_proxy_without_auth():
    config = {"username": "user", "password": "pass", "hostname": "proxy.lan"}
    formated_proxy = httpclient.format_proxy(config, auth=False)
    assert formated_proxy == "http://proxy.lan:80"


@pytest.mark.parametrize(
    "name,expected",
    [
        (None, r"^Mopidy/[^ ]+ CPython|/[^ ]+$"),
        ("Foo", r"^Foo Mopidy/[^ ]+ CPython|/[^ ]+$"),
        ("Foo/1.2.3", r"^Foo/1.2.3 Mopidy/[^ ]+ CPython|/[^ ]+$"),
    ],
)
def test_format_user_agent(name, expected):
    assert re.match(expected, httpclient.format_user_agent(name))



import logging

logger = logging.getLogger(__name__)

try:
    import dbus
except ImportError:
    dbus = None


# XXX: Hack to workaround introspection bug caused by gnome-keyring, should be
# fixed by version 3.5 per:
# https://git.gnome.org/browse/gnome-keyring/commit/?id=5dccbe88eb94eea9934e2b7
if dbus:
    EMPTY_STRING = dbus.String("", variant_level=1)
else:
    EMPTY_STRING = ""


FETCH_ERROR = (
    "Fetching passwords from your keyring failed. Any passwords "
    "stored in the keyring will not be available."
)


def fetch():
    if not dbus:
        logger.debug("%s (dbus not installed)", FETCH_ERROR)
        return []

    try:
        bus = dbus.SessionBus()
    except dbus.exceptions.DBusException as e:
        logger.debug("%s (%s)", FETCH_ERROR, e)
        return []

    if not bus.name_has_owner("org.freedesktop.secrets"):
        logger.debug(
            "%s (org.freedesktop.secrets service not running)", FETCH_ERROR
        )
        return []

    service = _service(bus)
    session = service.OpenSession("plain", EMPTY_STRING)[1]
    items, locked = service.SearchItems({"service": "mopidy"})

    if not locked and not items:
        return []

    if locked:
        # There is a chance we can unlock without prompting the users...
        items, prompt = service.Unlock(locked)
        if prompt != "/":
            _prompt(bus, prompt).Dismiss()
            logger.debug("%s (Keyring is locked)", FETCH_ERROR)
            return []

    result = []
    secrets = service.GetSecrets(items, session, byte_arrays=True)
    for item_path, values in secrets.items():
        session_path, parameters, value, content_type = values
        attrs = _item_attributes(bus, item_path)
        result.append((attrs["section"], attrs["key"], bytes(value)))
    return result


def set(section, key, value):
    """Store a secret config value for a given section/key.

    Indicates if storage failed or succeeded.
    """
    if not dbus:
        logger.debug(
            "Saving %s/%s to keyring failed. (dbus not installed)", section, key
        )
        return False

    try:
        bus = dbus.SessionBus()
    except dbus.exceptions.DBusException as e:
        logger.debug("Saving %s/%s to keyring failed. (%s)", section, key, e)
        return False

    if not bus.name_has_owner("org.freedesktop.secrets"):
        logger.debug(
            "Saving %s/%s to keyring failed. "
            "(org.freedesktop.secrets service not running)",
            section,
            key,
        )
        return False

    service = _service(bus)
    collection = _collection(bus)
    if not collection:
        return False

    if isinstance(value, str):
        value = value.encode()

    session = service.OpenSession("plain", EMPTY_STRING)[1]
    secret = dbus.Struct(
        (session, "", dbus.ByteArray(value), "plain/text; charset=utf8")
    )
    label = f"mopidy: {section}/{key}"
    attributes = {"service": "mopidy", "section": section, "key": key}
    properties = {
        "org.freedesktop.Secret.Item.Label": label,
        "org.freedesktop.Secret.Item.Attributes": attributes,
    }

    try:
        item, prompt = collection.CreateItem(properties, secret, True)
    except dbus.exceptions.DBusException as e:
        # TODO: catch IsLocked errors etc.
        logger.debug("Saving %s/%s to keyring failed. (%s)", section, key, e)
        return False

    if prompt == "/":
        return True

    _prompt(bus, prompt).Dismiss()
    logger.debug(
        "Saving secret %s/%s failed. (Keyring is locked)", section, key
    )
    return False


def _service(bus):
    return _interface(
        bus, "/org/freedesktop/secrets", "org.freedesktop.Secret.Service"
    )


# NOTE: depending on versions and setup 'default' might not exists, so try and
# use it but fall back to the 'login' collection, and finally the 'session' one
# if all else fails. We should probably create a keyring/collection setting
# that allows users to set this so they have control over where their secrets
# get stored.
def _collection(bus):
    for name in "aliases/default", "collection/login", "collection/session":
        path = "/org/freedesktop/secrets/" + name
        if _collection_exists(bus, path):
            break
    else:
        return None
    return _interface(bus, path, "org.freedesktop.Secret.Collection")


# NOTE: Hack to probe if a given collection actually exists. Needed to work
# around an introspection bug in setting passwords for non-existant aliases.
def _collection_exists(bus, path):
    try:
        item = _interface(bus, path, "org.freedesktop.DBus.Properties")
        item.Get("org.freedesktop.Secret.Collection", "Label")
        return True
    except dbus.exceptions.DBusException:
        return False


# NOTE: We could call prompt.Prompt('') to unlock the keyring when it is not
# '/', but we would then also have to arrange to setup signals to wait until
# this has been completed. So for now we just dismiss the prompt and expect
# keyrings to be unlocked.
def _prompt(bus, path):
    return _interface(bus, path, "Prompt")


def _item_attributes(bus, path):
    item = _interface(bus, path, "org.freedesktop.DBus.Properties")
    result = item.Get("org.freedesktop.Secret.Item", "Attributes")
    return {str(k): str(v) for k, v in result.items()}


def _interface(bus, path, interface):
    obj = bus.get_object("org.freedesktop.secrets", path)
    return dbus.Interface(obj, interface)



import unittest
from unittest import mock

import pykka

from mopidy import core
from mopidy.internal import deprecation
from mopidy.models import Track

from tests import dummy_backend


@mock.patch.object(core.CoreListener, "send")
class BackendEventsTest(unittest.TestCase):
    def setUp(self):  # noqa: N802
        config = {"core": {"max_tracklist_length": 10000}}

        self.backend = dummy_backend.create_proxy()
        self.backend.library.dummy_library = [
            Track(uri="dummy:a"),
            Track(uri="dummy:b"),
        ]

        with deprecation.ignore():
            self.core = core.Core.start(config, backends=[self.backend]).proxy()

    def tearDown(self):  # noqa: N802
        pykka.ActorRegistry.stop_all()

    def test_forwards_backend_playlists_loaded_event_to_frontends(self, send):
        self.core.playlists_loaded().get()

        assert send.call_args[0][0] == "playlists_loaded"

    def test_forwards_mixer_volume_changed_event_to_frontends(self, send):
        self.core.volume_changed(volume=60).get()

        assert send.call_args[0][0] == "volume_changed"
        assert send.call_args[1]["volume"] == 60

    def test_forwards_mixer_mute_changed_event_to_frontends(self, send):
        self.core.mute_changed(mute=True).get()

        assert send.call_args[0][0] == "mute_changed"
        assert send.call_args[1]["mute"] is True

    def test_tracklist_add_sends_tracklist_changed_event(self, send):
        self.core.tracklist.add(uris=["dummy:a"]).get()

        assert send.call_args[0][0] == "tracklist_changed"

    def test_tracklist_clear_sends_tracklist_changed_event(self, send):
        self.core.tracklist.add(uris=["dummy:a"]).get()

        self.core.tracklist.clear().get()

        assert send.call_args[0][0] == "tracklist_changed"

    def test_tracklist_move_sends_tracklist_changed_event(self, send):
        self.core.tracklist.add(uris=["dummy:a", "dummy:b"]).get()

        self.core.tracklist.move(0, 1, 1).get()

        assert send.call_args[0][0] == "tracklist_changed"

    def test_tracklist_remove_sends_tracklist_changed_event(self, send):
        self.core.tracklist.add(uris=["dummy:a"]).get()

        self.core.tracklist.remove({"uri": ["dummy:a"]}).get()

        assert send.call_args[0][0] == "tracklist_changed"

    def test_tracklist_shuffle_sends_tracklist_changed_event(self, send):
        self.core.tracklist.add(uris=["dummy:a", "dummy:b"]).get()

        self.core.tracklist.shuffle().get()

        assert send.call_args[0][0] == "tracklist_changed"

    def test_playlists_refresh_sends_playlists_loaded_event(self, send):
        self.core.playlists.refresh().get()

        assert send.call_args[0][0] == "playlists_loaded"

    def test_playlists_refresh_uri_sends_playlists_loaded_event(self, send):
        self.core.playlists.refresh(uri_scheme="dummy").get()

        assert send.call_args[0][0] == "playlists_loaded"

    def test_playlists_create_sends_playlist_changed_event(self, send):
        self.core.playlists.create("foo").get()

        assert send.call_args[0][0] == "playlist_changed"

    def test_playlists_delete_sends_playlist_deleted_event(self, send):
        playlist = self.core.playlists.create("foo").get()
        self.core.playlists.delete(playlist.uri).get()

        assert send.call_args[0][0] == "playlist_deleted"

    def test_playlists_save_sends_playlist_changed_event(self, send):
        playlist = self.core.playlists.create("foo").get()
        playlist = playlist.replace(name="bar")

        self.core.playlists.save(playlist).get()

        assert send.call_args[0][0] == "playlist_changed"



import unittest
from unittest import mock

from mopidy import config, ext

from tests import path_to_data_dir


class LoadConfigTest(unittest.TestCase):
    def test_load_nothing(self):
        assert {} == config._load([], [], [])

    def test_load_missing_file(self):
        file0 = path_to_data_dir("file0.conf")
        result = config._load([file0], [], [])
        assert {} == result

    @mock.patch("os.access")
    def test_load_nonreadable_file(self, access_mock):
        access_mock.return_value = False
        file1 = path_to_data_dir("file1.conf")
        result = config._load([file1], [], [])
        assert {} == result

    def test_load_single_default(self):
        default = b"[foo]\nbar = baz"
        expected = {"foo": {"bar": "baz"}}
        result = config._load([], [default], [])
        assert expected == result

    def test_unicode_default(self):
        default = "[foo]\nbar = æøå"
        expected = {"foo": {"bar": "æøå"}}
        result = config._load([], [default], [])
        assert expected == result

    def test_load_defaults(self):
        default1 = b"[foo]\nbar = baz"
        default2 = b"[foo2]\n"
        expected = {"foo": {"bar": "baz"}, "foo2": {}}
        result = config._load([], [default1, default2], [])
        assert expected == result

    def test_load_single_override(self):
        override = ("foo", "bar", "baz")
        expected = {"foo": {"bar": "baz"}}
        result = config._load([], [], [override])
        assert expected == result

    def test_load_overrides(self):
        override1 = ("foo", "bar", "baz")
        override2 = ("foo2", "bar", "baz")
        expected = {"foo": {"bar": "baz"}, "foo2": {"bar": "baz"}}
        result = config._load([], [], [override1, override2])
        assert expected == result

    def test_load_single_file(self):
        file1 = path_to_data_dir("file1.conf")
        expected = {"foo": {"bar": "baz"}}
        result = config._load([file1], [], [])
        assert expected == result

    def test_load_files(self):
        file1 = path_to_data_dir("file1.conf")
        file2 = path_to_data_dir("file2.conf")
        expected = {"foo": {"bar": "baz"}, "foo2": {"bar": "baz"}}
        result = config._load([file1, file2], [], [])
        assert expected == result

    def test_load_directory(self):
        directory = path_to_data_dir("conf1.d")
        expected = {"foo": {"bar": "baz"}, "foo2": {"bar": "baz"}}
        result = config._load([directory], [], [])
        assert expected == result

    def test_load_directory_only_conf_files(self):
        directory = path_to_data_dir("conf2.d")
        expected = {"foo": {"bar": "baz"}}
        result = config._load([directory], [], [])
        assert expected == result

    def test_load_file_with_utf8(self):
        expected = {"foo": {"bar": "æøå"}}
        result = config._load([path_to_data_dir("file3.conf")], [], [])
        assert expected == result

    def test_load_file_with_error(self):
        expected = {"foo": {"bar": "baz"}}
        result = config._load([path_to_data_dir("file4.conf")], [], [])
        assert expected == result


class ValidateTest(unittest.TestCase):
    def setUp(self):  # noqa: N802
        self.schema = config.ConfigSchema("foo")
        self.schema["bar"] = config.ConfigValue()

    def test_empty_config_no_schemas(self):
        conf, errors = config._validate({}, [])
        assert {} == conf
        assert {} == errors

    def test_config_no_schemas(self):
        raw_config = {"foo": {"bar": "baz"}}
        conf, errors = config._validate(raw_config, [])
        assert {} == conf
        assert {} == errors

    def test_empty_config_single_schema(self):
        conf, errors = config._validate({}, [self.schema])
        assert {"foo": {"bar": None}} == conf
        assert {"foo": {"bar": "config key not found."}} == errors

    def test_config_single_schema(self):
        raw_config = {"foo": {"bar": "baz"}}
        conf, errors = config._validate(raw_config, [self.schema])
        assert {"foo": {"bar": "baz"}} == conf
        assert {} == errors

    def test_config_single_schema_config_error(self):
        raw_config = {"foo": {"bar": "baz"}}
        self.schema["bar"] = mock.Mock()
        self.schema["bar"].deserialize.side_effect = ValueError("bad")
        conf, errors = config._validate(raw_config, [self.schema])
        assert {"foo": {"bar": None}} == conf
        assert {"foo": {"bar": "bad"}} == errors

    # TODO: add more tests


INPUT_CONFIG = """# comments before first section should work

[section] anything goes ; after the [] block it seems.
; this is a valid comment
this-should-equal-baz = baz ; as this is a comment
this-should-equal-everything = baz # as this is not a comment

# this is also a comment ; and the next line should be a blank comment.
;
# foo # = should all be treated as a comment."""

PROCESSED_CONFIG = """[__COMMENTS__]
__HASH0__ = comments before first section should work
__BLANK1__ =
[section]
__SECTION2__ = anything goes
__INLINE3__ = after the [] block it seems.
__SEMICOLON4__ = this is a valid comment
this-should-equal-baz = baz
__INLINE5__ = as this is a comment
this-should-equal-everything = baz # as this is not a comment
__BLANK6__ =
__HASH7__ = this is also a comment
__INLINE8__ = and the next line should be a blank comment.
__SEMICOLON9__ =
__HASH10__ = foo # = should all be treated as a comment."""


class PreProcessorTest(unittest.TestCase):
    maxDiff = None  # Show entire diff.  # noqa: N815

    def test_empty_config(self):
        result = config._preprocess("")
        assert result == "[__COMMENTS__]"

    def test_plain_section(self):
        result = config._preprocess("[section]\nfoo = bar")
        assert result == "[__COMMENTS__]\n[section]\nfoo = bar"

    def test_initial_comments(self):
        result = config._preprocess("; foobar")
        assert result == "[__COMMENTS__]\n__SEMICOLON0__ = foobar"

        result = config._preprocess("# foobar")
        assert result == "[__COMMENTS__]\n__HASH0__ = foobar"

        result = config._preprocess("; foo\n# bar")
        assert result == "[__COMMENTS__]\n__SEMICOLON0__ = foo\n__HASH1__ = bar"

    def test_initial_comment_inline_handling(self):
        result = config._preprocess("; foo ; bar ; baz")
        assert result == (
            "[__COMMENTS__]\n__SEMICOLON0__ = foo\n"
            "__INLINE1__ = bar\n__INLINE2__ = baz"
        )

    def test_inline_semicolon_comment(self):
        result = config._preprocess("[section]\nfoo = bar ; baz")
        assert (
            result == "[__COMMENTS__]\n[section]\nfoo = bar\n__INLINE0__ = baz"
        )

    def test_no_inline_hash_comment(self):
        result = config._preprocess("[section]\nfoo = bar # baz")
        assert result == "[__COMMENTS__]\n[section]\nfoo = bar # baz"

    def test_section_extra_text(self):
        result = config._preprocess("[section] foobar")
        assert result == "[__COMMENTS__]\n[section]\n__SECTION0__ = foobar"

    def test_section_extra_text_inline_semicolon(self):
        result = config._preprocess("[section] foobar ; baz")
        assert (
            result
            == "[__COMMENTS__]\n[section]\n__SECTION0__ = foobar\n__INLINE1__ = baz"
        )

    def test_conversion(self):
        """Tests all of the above cases at once."""
        result = config._preprocess(INPUT_CONFIG)
        assert result == PROCESSED_CONFIG


class PostProcessorTest(unittest.TestCase):
    maxDiff = None  # Show entire diff.  # noqa: N815

    def test_empty_config(self):
        result = config._postprocess("[__COMMENTS__]")
        assert result == ""

    def test_plain_section(self):
        result = config._postprocess(
            "[__COMMENTS__]\n" "[section]\n" "foo = bar"
        )
        assert result == "[section]\nfoo = bar"

    def test_initial_comments(self):
        result = config._postprocess(
            "[__COMMENTS__]\n" "__SEMICOLON0__ = foobar"
        )
        assert result == "; foobar"

        result = config._postprocess("[__COMMENTS__]\n" "__HASH0__ = foobar")
        assert result == "# foobar"

        result = config._postprocess(
            "[__COMMENTS__]\n" "__SEMICOLON0__ = foo\n" "__HASH1__ = bar"
        )
        assert result == "; foo\n# bar"

    def test_initial_comment_inline_handling(self):
        result = config._postprocess(
            "[__COMMENTS__]\n"
            "__SEMICOLON0__ = foo\n"
            "__INLINE1__ = bar\n"
            "__INLINE2__ = baz"
        )
        assert result == "; foo ; bar ; baz"

    def test_inline_semicolon_comment(self):
        result = config._postprocess(
            "[__COMMENTS__]\n" "[section]\n" "foo = bar\n" "__INLINE0__ = baz"
        )
        assert result == "[section]\nfoo = bar ; baz"

    def test_no_inline_hash_comment(self):
        result = config._preprocess("[section]\nfoo = bar # baz")
        assert result == "[__COMMENTS__]\n[section]\nfoo = bar # baz"

    def test_section_extra_text(self):
        result = config._postprocess(
            "[__COMMENTS__]\n" "[section]\n" "__SECTION0__ = foobar"
        )
        assert result == "[section] foobar"

    def test_section_extra_text_inline_semicolon(self):
        result = config._postprocess(
            "[__COMMENTS__]\n"
            "[section]\n"
            "__SECTION0__ = foobar\n"
            "__INLINE1__ = baz"
        )
        assert result == "[section] foobar ; baz"

    def test_conversion(self):
        result = config._postprocess(PROCESSED_CONFIG)
        assert result == INPUT_CONFIG


def test_format_initial():
    extension = ext.Extension()
    extension.ext_name = "foo"
    extension.get_default_config = lambda: None
    extensions_data = [
        ext.ExtensionData(
            extension=extension,
            entry_point=None,
            config_schema=None,
            config_defaults=None,
            command=None,
        ),
    ]

    result = config.format_initial(extensions_data)

    assert "# For further information" in result
    assert "[foo]\n" in result



import configparser
import itertools
import logging
import os
import pathlib
import re
from collections.abc import Mapping

from mopidy.config import keyring
from mopidy.config.schemas import ConfigSchema, MapConfigSchema
from mopidy.config.types import (
    Boolean,
    ConfigValue,
    Deprecated,
    DeprecatedValue,
    Hostname,
    Integer,
    List,
    LogColor,
    LogLevel,
    Path,
    Port,
    Secret,
    String,
)
from mopidy.internal import path, versioning

__all__ = [
    # TODO List everything that is reexported, not just the unused parts.
    "ConfigValue",
    "List",
]

logger = logging.getLogger(__name__)

_core_schema = ConfigSchema("core")
_core_schema["cache_dir"] = Path()
_core_schema["config_dir"] = Path()
_core_schema["data_dir"] = Path()
# MPD supports at most 10k tracks, some clients segfault when this is exceeded.
_core_schema["max_tracklist_length"] = Integer(minimum=1)
_core_schema["restore_state"] = Boolean(optional=True)

_logging_schema = ConfigSchema("logging")
_logging_schema["verbosity"] = Integer(minimum=-1, maximum=4)
_logging_schema["format"] = String()
_logging_schema["color"] = Boolean()
_logging_schema["console_format"] = Deprecated()
_logging_schema["debug_format"] = Deprecated()
_logging_schema["debug_file"] = Deprecated()
_logging_schema["config_file"] = Path(optional=True)

_loglevels_schema = MapConfigSchema("loglevels", LogLevel())
_logcolors_schema = MapConfigSchema("logcolors", LogColor())

_audio_schema = ConfigSchema("audio")
_audio_schema["mixer"] = String()
_audio_schema["mixer_track"] = Deprecated()
_audio_schema["mixer_volume"] = Integer(optional=True, minimum=0, maximum=100)
_audio_schema["output"] = String()
_audio_schema["visualizer"] = Deprecated()
_audio_schema["buffer_time"] = Integer(optional=True, minimum=1)

_proxy_schema = ConfigSchema("proxy")
_proxy_schema["scheme"] = String(
    optional=True, choices=["http", "https", "socks4", "socks5"]
)
_proxy_schema["hostname"] = Hostname(optional=True)
_proxy_schema["port"] = Port(optional=True)
_proxy_schema["username"] = String(optional=True)
_proxy_schema["password"] = Secret(optional=True)

# NOTE: if multiple outputs ever comes something like LogLevelConfigSchema
# _outputs_schema = config.AudioOutputConfigSchema()

_schemas = [
    _core_schema,
    _logging_schema,
    _loglevels_schema,
    _logcolors_schema,
    _audio_schema,
    _proxy_schema,
]

_INITIAL_HELP = """
# For further information about options in this file see:
#   https://docs.mopidy.com/
#
# The initial commented out values reflect the defaults as of:
#   {versions}
#
# Available options and defaults might have changed since then,
# run `mopidy config` to see the current effective config and
# `mopidy --version` to check the current version.
"""


def read(config_file):
    """Helper to load config defaults in same way across core and extensions"""
    return pathlib.Path(config_file).read_text(errors="surrogateescape")


def load(files, ext_schemas, ext_defaults, overrides):
    config_dir = pathlib.Path(__file__).parent
    defaults = [read(config_dir / "default.conf")]
    defaults.extend(ext_defaults)
    raw_config = _load(files, defaults, keyring.fetch() + (overrides or []))

    schemas = _schemas[:]
    schemas.extend(ext_schemas)
    return _validate(raw_config, schemas)


def format(config, ext_schemas, comments=None, display=True):
    schemas = _schemas[:]
    schemas.extend(ext_schemas)
    return _format(config, comments or {}, schemas, display, False)


def format_initial(extensions_data):
    config_dir = pathlib.Path(__file__).parent
    defaults = [read(config_dir / "default.conf")]
    defaults.extend(d.extension.get_default_config() for d in extensions_data)
    raw_config = _load([], defaults, [])

    schemas = _schemas[:]
    schemas.extend(d.extension.get_config_schema() for d in extensions_data)

    config, errors = _validate(raw_config, schemas)

    versions = [f"Mopidy {versioning.get_version()}"]
    extensions_data = sorted(
        extensions_data, key=lambda d: d.extension.dist_name
    )
    for data in extensions_data:
        versions.append(f"{data.extension.dist_name} {data.extension.version}")

    header = _INITIAL_HELP.strip().format(versions="\n#   ".join(versions))
    formatted_config = _format(
        config=config, comments={}, schemas=schemas, display=False, disable=True
    )
    return header + "\n\n" + formatted_config


def _load(files, defaults, overrides):
    parser = configparser.RawConfigParser()

    # TODO: simply return path to config file for defaults so we can load it
    # all in the same way?
    logger.info("Loading config from builtin defaults")
    for default in defaults:
        if isinstance(default, bytes):
            default = default.decode()
        parser.read_string(default)

    # Load config from a series of config files
    for f in files:
        f = path.expand_path(f)
        if f.is_dir():
            for g in f.iterdir():
                if g.is_file() and g.suffix == ".conf":
                    _load_file(parser, g.resolve())
        else:
            _load_file(parser, f.resolve())

    raw_config = {}
    for section in parser.sections():
        raw_config[section] = dict(parser.items(section))

    logger.info("Loading config from command line options")
    for section, key, value in overrides:
        raw_config.setdefault(section, {})[key] = value

    return raw_config


def _load_file(parser, file_path):
    if not file_path.exists():
        logger.debug(
            f"Loading config from {file_path.as_uri()} failed; "
            f"it does not exist"
        )
        return
    if not os.access(str(file_path), os.R_OK):
        logger.warning(
            f"Loading config from {file_path.as_uri()} failed; "
            f"read permission missing"
        )
        return

    try:
        logger.info(f"Loading config from {file_path.as_uri()}")
        with file_path.open("r") as fh:
            parser.read_file(fh)
    except configparser.MissingSectionHeaderError:
        logger.warning(
            f"Loading config from {file_path.as_uri()} failed; "
            f"it does not have a config section"
        )
    except configparser.ParsingError as e:
        linenos = ", ".join(str(lineno) for lineno, line in e.errors)
        logger.warning(
            f"Config file {file_path.as_uri()} has errors; "
            f"line {linenos} has been ignored"
        )
    except OSError:
        # TODO: if this is the initial load of logging config we might not
        # have a logger at this point, we might want to handle this better.
        logger.debug(f"Config file {file_path.as_uri()} not found; skipping")


def _validate(raw_config, schemas):
    # Get validated config
    config = {}
    errors = {}
    sections = set(raw_config)
    for schema in schemas:
        sections.discard(schema.name)
        values = raw_config.get(schema.name, {})
        result, error = schema.deserialize(values)
        if error:
            errors[schema.name] = error
        if result:
            config[schema.name] = result

    for section in sections:
        logger.warning(
            f"Ignoring config section {section!r} "
            f"because no matching extension was found"
        )

    return config, errors


def _format(config, comments, schemas, display, disable):
    output = []
    for schema in schemas:
        serialized = schema.serialize(
            config.get(schema.name, {}), display=display
        )
        if not serialized:
            continue
        output.append(f"[{schema.name}]")
        for key, value in serialized.items():
            if isinstance(value, DeprecatedValue):
                continue
            comment = comments.get(schema.name, {}).get(key, "")
            output.append(f"{key} =")
            if value is not None:
                output[-1] += " " + value
            if comment:
                output[-1] += "  ; " + comment.capitalize()
            if disable:
                output[-1] = re.sub(r"^", "#", output[-1], flags=re.M)
        output.append("")
    return "\n".join(output).strip()


def _preprocess(config_string):
    """Convert a raw config into a form that preserves comments etc."""
    results = ["[__COMMENTS__]"]
    counter = itertools.count(0)

    section_re = re.compile(r"^(\[[^\]]+\])\s*(.+)$")
    blank_line_re = re.compile(r"^\s*$")
    comment_re = re.compile(r"^(#|;)")
    inline_comment_re = re.compile(r" ;")

    def newlines(match):
        return f"__BLANK{next(counter):d}__ ="

    def comments(match):
        if match.group(1) == "#":
            return f"__HASH{next(counter):d}__ ="
        elif match.group(1) == ";":
            return f"__SEMICOLON{next(counter):d}__ ="

    def inlinecomments(match):
        return f"\n__INLINE{next(counter):d}__ ="

    def sections(match):
        return (
            f"{match.group(1)}\n__SECTION{next(counter):d}__ = {match.group(2)}"
        )

    for line in config_string.splitlines():
        line = blank_line_re.sub(newlines, line)
        line = section_re.sub(sections, line)
        line = comment_re.sub(comments, line)
        line = inline_comment_re.sub(inlinecomments, line)
        results.append(line)
    return "\n".join(results)


def _postprocess(config_string):
    """Converts a preprocessed config back to original form."""
    flags = re.IGNORECASE | re.MULTILINE
    result = re.sub(r"^\[__COMMENTS__\](\n|$)", "", config_string, flags=flags)
    result = re.sub(r"\n__INLINE\d+__ =(.*)$", r" ;\g<1>", result, flags=flags)
    result = re.sub(r"^__HASH\d+__ =(.*)$", r"#\g<1>", result, flags=flags)
    result = re.sub(r"^__SEMICOLON\d+__ =(.*)$", r";\g<1>", result, flags=flags)
    result = re.sub(r"\n__SECTION\d+__ =(.*)$", r"\g<1>", result, flags=flags)
    result = re.sub(r"^__BLANK\d+__ =$", "", result, flags=flags)
    return result


class Proxy(Mapping):
    def __init__(self, data):
        self._data = data

    def __getitem__(self, key):
        item = self._data.__getitem__(key)
        if isinstance(item, dict):
            return Proxy(item)
        return item

    def __iter__(self):
        return self._data.__iter__()

    def __len__(self):
        return self._data.__len__()

    def __repr__(self):
        return f"Proxy({self._data!r})"



import logging
import signal
import sys

import pykka.debug

from mopidy import commands
from mopidy import config as config_lib
from mopidy import ext
from mopidy.internal import log, path, process, versioning
from mopidy.internal.gi import Gst  # noqa: F401

try:
    # Make GLib's mainloop the event loop for python-dbus
    import dbus.mainloop.glib

    dbus.mainloop.glib.threads_init()
    dbus.mainloop.glib.DBusGMainLoop(set_as_default=True)
except ImportError:
    pass


logger = logging.getLogger(__name__)


def main():
    log.bootstrap_delayed_logging()
    logger.info(f"Starting Mopidy {versioning.get_version()}")

    signal.signal(signal.SIGTERM, process.sigterm_handler)
    # Windows does not have signal.SIGUSR1
    if hasattr(signal, "SIGUSR1"):
        signal.signal(signal.SIGUSR1, pykka.debug.log_thread_tracebacks)

    try:
        registry = ext.Registry()

        root_cmd = commands.RootCommand()
        config_cmd = commands.ConfigCommand()
        deps_cmd = commands.DepsCommand()

        root_cmd.set(extension=None, registry=registry)
        root_cmd.add_child("config", config_cmd)
        root_cmd.add_child("deps", deps_cmd)

        extensions_data = ext.load_extensions()

        for data in extensions_data:
            if data.command:  # TODO: check isinstance?
                data.command.set(extension=data.extension)
                root_cmd.add_child(data.extension.ext_name, data.command)

        args = root_cmd.parse(sys.argv[1:])

        config, config_errors = config_lib.load(
            args.config_files,
            [d.config_schema for d in extensions_data],
            [d.config_defaults for d in extensions_data],
            args.config_overrides,
        )

        create_core_dirs(config)
        create_initial_config_file(args, extensions_data)

        log.setup_logging(
            config, args.base_verbosity_level, args.verbosity_level
        )

        extensions = {
            "validate": [],
            "config": [],
            "disabled": [],
            "enabled": [],
        }
        for data in extensions_data:
            extension = data.extension

            # TODO: factor out all of this to a helper that can be tested
            if not ext.validate_extension_data(data):
                config[extension.ext_name] = {"enabled": False}
                config_errors[extension.ext_name] = {
                    "enabled": "extension disabled by self check."
                }
                extensions["validate"].append(extension)
            elif not config[extension.ext_name]["enabled"]:
                config[extension.ext_name] = {"enabled": False}
                config_errors[extension.ext_name] = {
                    "enabled": "extension disabled by user config."
                }
                extensions["disabled"].append(extension)
            elif config_errors.get(extension.ext_name):
                config[extension.ext_name]["enabled"] = False
                config_errors[extension.ext_name][
                    "enabled"
                ] = "extension disabled due to config errors."
                extensions["config"].append(extension)
            else:
                extensions["enabled"].append(extension)

        log_extension_info(
            [d.extension for d in extensions_data], extensions["enabled"]
        )

        # Config and deps commands are simply special cased for now.
        if args.command == config_cmd:
            schemas = [d.config_schema for d in extensions_data]
            return args.command.run(config, config_errors, schemas)
        elif args.command == deps_cmd:
            return args.command.run()

        check_config_errors(config, config_errors, extensions)

        if not extensions["enabled"]:
            logger.error("No extension enabled, exiting...")
            sys.exit(1)

        # Read-only config from here on, please.
        proxied_config = config_lib.Proxy(config)

        if args.extension and args.extension not in extensions["enabled"]:
            logger.error(
                "Unable to run command provided by disabled extension %s",
                args.extension.ext_name,
            )
            return 1

        for extension in extensions["enabled"]:
            try:
                extension.setup(registry)
            except Exception:
                # TODO: would be nice a transactional registry. But sadly this
                # is a bit tricky since our current API is giving out a mutable
                # list. We might however be able to replace this with a
                # collections.Sequence to provide a RO view.
                logger.exception(
                    f"Extension {extension.ext_name} failed during setup. "
                    f"This might have left the registry in a bad state."
                )

        # Anything that wants to exit after this point must use
        # mopidy.internal.process.exit_process as actors can have been started.
        try:
            return args.command.run(args, proxied_config)
        except NotImplementedError:
            print(root_cmd.format_help())
            return 1

    except KeyboardInterrupt:
        pass
    except Exception as ex:
        logger.exception(ex)
        raise


def create_core_dirs(config):
    path.get_or_create_dir(config["core"]["cache_dir"])
    path.get_or_create_dir(config["core"]["config_dir"])
    path.get_or_create_dir(config["core"]["data_dir"])


def create_initial_config_file(args, extensions_data):
    """Initialize whatever the last config file is with defaults"""

    config_file = path.expand_path(args.config_files[-1])

    if config_file.exists():
        return

    try:
        default = config_lib.format_initial(extensions_data)
        path.get_or_create_file(
            config_file,
            mkdir=False,
            content=default.encode(errors="surrogateescape"),
        )
        logger.info(f"Initialized {config_file.as_uri()} with default config")
    except OSError as exc:
        logger.warning(
            f"Unable to initialize {config_file.as_uri()} with default config: {exc}"
        )


def log_extension_info(all_extensions, enabled_extensions):
    # TODO: distinguish disabled vs blocked by env?
    enabled_names = {e.ext_name for e in enabled_extensions}
    disabled_names = {e.ext_name for e in all_extensions} - enabled_names
    logger.info("Enabled extensions: %s", ", ".join(enabled_names) or "none")
    logger.info("Disabled extensions: %s", ", ".join(disabled_names) or "none")


def check_config_errors(config, errors, extensions):
    fatal_errors = []
    extension_names = {}
    all_extension_names = set()

    for state in extensions:
        extension_names[state] = {e.ext_name for e in extensions[state]}
        all_extension_names.update(extension_names[state])

    for section in sorted(errors):
        if not errors[section]:
            continue

        if section not in all_extension_names:
            logger.warning(f"Found fatal {section} configuration errors:")
            fatal_errors.append(section)
        elif section in extension_names["config"]:
            del errors[section]["enabled"]
            logger.warning(
                f"Found {section} configuration errors. "
                f"The extension has been automatically disabled:"
            )
        else:
            continue

        for field, msg in errors[section].items():
            logger.warning(f"  {section}/{field} {msg}")

    if extensions["config"]:
        logger.warning(
            "Please fix the extension configuration errors or "
            "disable the extensions to silence these messages."
        )

    if fatal_errors:
        logger.error("Please fix fatal configuration errors, exiting...")
        sys.exit(1)


if __name__ == "__main__":
    sys.exit(main())



import json
import unittest
from unittest import mock

from mopidy.http import actor


@mock.patch("mopidy.http.handlers.WebSocketHandler.broadcast")
class HttpEventsTest(unittest.TestCase):
    def setUp(self):
        self.io_loop = mock.Mock()

    def test_track_playback_paused_is_broadcasted(self, broadcast):
        actor.on_event("track_playback_paused", self.io_loop, foo="bar")

        self.assertDictEqual(
            json.loads(str(broadcast.call_args[0][0])),
            {"event": "track_playback_paused", "foo": "bar"},
        )

    def test_track_playback_resumed_is_broadcasted(self, broadcast):
        actor.on_event("track_playback_resumed", self.io_loop, foo="bar")

        self.assertDictEqual(
            json.loads(str(broadcast.call_args[0][0])),
            {"event": "track_playback_resumed", "foo": "bar"},
        )



import fnmatch
import logging
import re
import time
import urllib

import pykka

from mopidy import audio as audio_lib
from mopidy import backend, exceptions, stream
from mopidy.audio import scan, tags
from mopidy.internal import http, playlists
from mopidy.models import Track

logger = logging.getLogger(__name__)


class StreamBackend(pykka.ThreadingActor, backend.Backend):
    def __init__(self, config, audio):
        super().__init__()

        self._scanner = scan.Scanner(
            timeout=config["stream"]["timeout"], proxy_config=config["proxy"]
        )

        self._session = http.get_requests_session(
            proxy_config=config["proxy"],
            user_agent=(
                f"{stream.Extension.dist_name}/{stream.Extension.version}"
            ),
        )

        blacklist = config["stream"]["metadata_blacklist"]
        self._blacklist_re = re.compile(
            r"^(%s)$" % "|".join(fnmatch.translate(u) for u in blacklist)
        )

        self._timeout = config["stream"]["timeout"]

        self.library = StreamLibraryProvider(backend=self)
        self.playback = StreamPlaybackProvider(audio=audio, backend=self)
        self.playlists = None

        self.uri_schemes = audio_lib.supported_uri_schemes(
            config["stream"]["protocols"]
        )

        if "file" in self.uri_schemes and config["file"]["enabled"]:
            logger.warning(
                'The stream/protocols config value includes the "file" '
                'protocol. "file" playback is now handled by Mopidy-File. '
                "Please remove it from the stream/protocols config."
            )
            self.uri_schemes -= {"file"}


class StreamLibraryProvider(backend.LibraryProvider):
    def lookup(self, uri):
        if urllib.parse.urlsplit(uri).scheme not in self.backend.uri_schemes:
            return []

        if self.backend._blacklist_re.match(uri):
            logger.debug("URI matched metadata lookup blacklist: %s", uri)
            return [Track(uri=uri)]

        _, scan_result = _unwrap_stream(
            uri,
            timeout=self.backend._timeout,
            scanner=self.backend._scanner,
            requests_session=self.backend._session,
        )

        if scan_result:
            track = tags.convert_tags_to_track(scan_result.tags).replace(
                uri=uri, length=scan_result.duration
            )
        else:
            logger.warning("Problem looking up %s", uri)
            track = Track(uri=uri)

        return [track]


class StreamPlaybackProvider(backend.PlaybackProvider):
    def translate_uri(self, uri):
        if urllib.parse.urlsplit(uri).scheme not in self.backend.uri_schemes:
            return None

        if self.backend._blacklist_re.match(uri):
            logger.debug("URI matched metadata lookup blacklist: %s", uri)
            return uri

        unwrapped_uri, _ = _unwrap_stream(
            uri,
            timeout=self.backend._timeout,
            scanner=self.backend._scanner,
            requests_session=self.backend._session,
        )
        return unwrapped_uri


# TODO: cleanup the return value of this.
def _unwrap_stream(uri, timeout, scanner, requests_session):
    """
    Get a stream URI from a playlist URI, ``uri``.

    Unwraps nested playlists until something that's not a playlist is found or
    the ``timeout`` is reached.
    """

    original_uri = uri
    seen_uris = set()
    deadline = time.time() + timeout

    while time.time() < deadline:
        if uri in seen_uris:
            logger.info(
                "Unwrapping stream from URI (%s) failed: "
                "playlist referenced itself",
                uri,
            )
            return None, None
        else:
            seen_uris.add(uri)

        logger.debug("Unwrapping stream from URI: %s", uri)

        try:
            scan_timeout = deadline - time.time()
            if scan_timeout < 0:
                logger.info(
                    "Unwrapping stream from URI (%s) failed: "
                    "timed out in %sms",
                    uri,
                    timeout,
                )
                return None, None
            scan_result = scanner.scan(uri, timeout=scan_timeout)
        except exceptions.ScannerError as exc:
            logger.debug("GStreamer failed scanning URI (%s): %s", uri, exc)
            scan_result = None

        if scan_result is not None:
            has_interesting_mime = (
                scan_result.mime is not None
                and not scan_result.mime.startswith("text/")
                and not scan_result.mime.startswith("application/")
            )
            if scan_result.playable or has_interesting_mime:
                logger.debug(
                    "Unwrapped potential %s stream: %s", scan_result.mime, uri
                )
                return uri, scan_result

        download_timeout = deadline - time.time()
        if download_timeout < 0:
            logger.info(
                "Unwrapping stream from URI (%s) failed: timed out in %sms",
                uri,
                timeout,
            )
            return None, None
        content = http.download(
            requests_session, uri, timeout=download_timeout / 1000
        )

        if content is None:
            logger.info(
                "Unwrapping stream from URI (%s) failed: "
                "error downloading URI %s",
                original_uri,
                uri,
            )
            return None, None

        uris = playlists.parse(content)
        if not uris:
            logger.debug(
                "Failed parsing URI (%s) as playlist; found potential stream.",
                uri,
            )
            return uri, None

        # TODO Test streams and return first that seems to be playable
        new_uri = uris[0]
        logger.debug("Parsed playlist (%s) and found new URI: %s", uri, new_uri)
        uri = urllib.parse.urljoin(uri, new_uri)



import io
import pathlib

import pytest

from mopidy.m3u import translator
from mopidy.m3u.translator import path_to_uri
from mopidy.models import Playlist, Ref, Track


def loads(s, basedir):
    return translator.load_items(io.StringIO(s), basedir)


def dumps(items):
    fp = io.StringIO()
    translator.dump_items(items, fp)
    return fp.getvalue()


@pytest.mark.parametrize(
    "path,scheme,expected",
    [
        ("test", None, "m3u:test"),
        ("test.m3u", None, "m3u:test.m3u"),
        ("./test.m3u", None, "m3u:test.m3u"),
        ("foo/../test.m3u", None, "m3u:test.m3u"),
        ("Test Playlist.m3u", None, "m3u:Test%20Playlist.m3u"),
        ("test.mp3", "file", "file:///test.mp3"),
    ],
)
def test_path_to_uri(path, scheme, expected):
    if scheme is not None:
        assert path_to_uri(pathlib.Path(path), scheme) == expected
    else:
        assert path_to_uri(pathlib.Path(path)) == expected


def test_latin1_path_to_uri():
    bytes_path = "æøå.m3u".encode("latin-1")
    path = pathlib.Path(bytes_path.decode(errors="surrogateescape"))

    assert translator.path_to_uri(path) == "m3u:%E6%F8%E5.m3u"


def test_utf8_path_to_uri():
    bytes_path = "æøå.m3u".encode()
    path = pathlib.Path(bytes_path.decode())

    assert translator.path_to_uri(path) == "m3u:%C3%A6%C3%B8%C3%A5.m3u"


@pytest.mark.parametrize(
    "path,expected",
    [
        ("test", "test"),
        ("test.m3u", "test"),
        ("../test.m3u", "test"),
        ("testæ.m3u", "testæ"),
    ],
)
def test_name_from_path(path, expected):
    assert translator.name_from_path(pathlib.Path(path)) == expected


def test_path_from_name():
    from mopidy.m3u.translator import path_from_name

    assert path_from_name("test") == pathlib.Path("test")
    assert path_from_name("test", ".m3u") == pathlib.Path("test.m3u")
    assert path_from_name("foo/bar", sep="-") == pathlib.Path("foo-bar")


@pytest.mark.parametrize(
    "path,expected",
    [
        ("test.m3u", ("m3u:test.m3u", "test")),
        ("Test Playlist.m3u", ("m3u:Test%20Playlist.m3u", "Test Playlist")),
    ],
)
def test_path_to_ref(path, expected):
    from mopidy.m3u.translator import path_to_ref

    result = path_to_ref(pathlib.Path(path))
    assert Ref.playlist(uri=expected[0], name=expected[1]) == result


@pytest.mark.parametrize(
    "contents,basedir,expected",
    [
        ("", ".", None),
        ("test.mp3", "/playlists", ("file:///playlists/test.mp3", "test")),
        ("../test.mp3", "/playlists", ("file:///test.mp3", "test")),
        ("/test.mp3", ".", ("file:///test.mp3", "test")),
        ("file:///test.mp3", ".", ("file:///test.mp3", None)),
        ("http://example.com/stream", ".", ("http://example.com/stream", None)),
        (
            "#EXTM3U\n#EXTINF:42,Test\nfile:///test.mp3\n",
            ".",
            ("file:///test.mp3", "Test"),
        ),
        (
            "#EXTM3U\n#EXTINF:-1,Test\nhttp://example.com/stream\n",
            ".",
            ("http://example.com/stream", "Test"),
        ),
    ],
)
def test_load_items(contents, basedir, expected):
    result = loads(contents, pathlib.Path(basedir))
    if expected is not None:
        assert [Ref.track(uri=expected[0], name=expected[1])] == result
    else:
        assert [] == result


def test_dump_items():
    assert dumps([]) == ""
    assert dumps([Ref.track(uri="file:///test.mp3")]) == ("file:///test.mp3\n")
    assert dumps([Ref.track(uri="file:///test.mp3", name="test")]) == (
        "#EXTM3U\n" "#EXTINF:-1,test\n" "file:///test.mp3\n"
    )
    assert dumps([Track(uri="file:///test.mp3", name="test", length=42)]) == (
        "#EXTM3U\n" "#EXTINF:-1,test\n" "file:///test.mp3\n"
    )
    assert dumps([Track(uri="http://example.com/stream")]) == (
        "http://example.com/stream\n"
    )
    assert dumps([Track(uri="http://example.com/stream", name="Test")]) == (
        "#EXTM3U\n" "#EXTINF:-1,Test\n" "http://example.com/stream\n"
    )


def test_playlist():
    from mopidy.m3u.translator import playlist

    path = pathlib.Path("test.m3u")

    assert playlist(path) == Playlist(uri="m3u:test.m3u", name="test")
    assert playlist(path, [Ref(uri="file:///test.mp3")], 1) == Playlist(
        uri="m3u:test.m3u",
        name="test",
        tracks=[Track(uri="file:///test.mp3")],
        last_modified=1000,
    )



import unittest
from unittest.mock import Mock, patch

from mopidy.internal import network


class TryIPv6SocketTest(unittest.TestCase):
    @patch("socket.has_ipv6", False)
    def test_system_that_claims_no_ipv6_support(self):
        assert not network.try_ipv6_socket()

    @patch("socket.has_ipv6", True)
    @patch("socket.socket")
    def test_system_with_broken_ipv6(self, socket_mock):
        socket_mock.side_effect = IOError()
        assert not network.try_ipv6_socket()

    @patch("socket.has_ipv6", True)
    @patch("socket.socket")
    def test_with_working_ipv6(self, socket_mock):
        socket_mock.return_value = Mock()
        assert network.try_ipv6_socket()


class FormatHostnameTest(unittest.TestCase):
    @patch("mopidy.internal.network.has_ipv6", True)
    def test_format_hostname_prefixes_ipv4_addresses_when_ipv6_available(self):
        network.has_ipv6 = True
        assert network.format_hostname("0.0.0.0") == "::ffff:0.0.0.0"
        assert network.format_hostname("1.0.0.1") == "::ffff:1.0.0.1"

    @patch("mopidy.internal.network.has_ipv6", False)
    def test_format_hostname_does_nothing_when_only_ipv4_available(self):
        network.has_ipv6 = False
        assert network.format_hostname("0.0.0.0") == "0.0.0.0"



# flake8: noqa
from .actor import Audio
from .constants import PlaybackState
from .listener import AudioListener
from .utils import (
    calculate_duration,
    create_buffer,
    millisecond_to_clocktime,
    supported_uri_schemes,
)



import logging
import pathlib
import re
import urllib

from mopidy.internal import xdg

logger = logging.getLogger(__name__)


XDG_DIRS = xdg.get_dirs()


def get_or_create_dir(dir_path):
    dir_path = expand_path(dir_path)
    if dir_path.is_file():
        raise OSError(
            f"A file with the same name as the desired dir, "
            f"{dir_path!r}, already exists."
        )
    elif not dir_path.is_dir():
        logger.info(f"Creating dir {dir_path.as_uri()}")
        dir_path.mkdir(mode=0o755, parents=True)
    return dir_path


def get_or_create_file(file_path, mkdir=True, content=None):
    file_path = expand_path(file_path)
    if isinstance(content, str):
        content = content.encode()
    if mkdir:
        get_or_create_dir(file_path.parent)
    if not file_path.is_file():
        logger.info(f"Creating file {file_path.as_uri()}")
        file_path.touch(exist_ok=False)
        if content is not None:
            file_path.write_bytes(content)
    return file_path


def get_unix_socket_path(socket_path):
    match = re.search("^unix:(.*)", socket_path)
    if not match:
        return None
    return match.group(1)


def path_to_uri(path):
    """
    Convert OS specific path to file:// URI.

    Accepts either unicode strings or bytestrings. The encoding of any
    bytestring will be maintained so that :func:`uri_to_path` can return the
    same bytestring.

    Returns a file:// URI as an unicode string.
    """
    return pathlib.Path(path).as_uri()


def uri_to_path(uri):
    """
    Convert an URI to a OS specific path.
    """
    bytes_path = urllib.parse.unquote_to_bytes(urllib.parse.urlsplit(uri).path)
    unicode_path = bytes_path.decode(errors="surrogateescape")
    return pathlib.Path(unicode_path)


def expand_path(path):
    if isinstance(path, bytes):
        path = path.decode(errors="surrogateescape")
    path = str(pathlib.Path(path))

    for xdg_var, xdg_dir in XDG_DIRS.items():
        path = path.replace("$" + xdg_var, str(xdg_dir))
    if "$" in path:
        return None

    return pathlib.Path(path).expanduser().resolve()


def is_path_inside_base_dir(path, base_path):
    if isinstance(path, bytes):
        path = path.decode(errors="surrogateescape")
    if isinstance(base_path, bytes):
        base_path = base_path.decode(errors="surrogateescape")

    path = pathlib.Path(path).resolve()
    base_path = pathlib.Path(base_path).resolve()

    if path.is_file():
        # Use dir of file for prefix comparision, so we don't accept
        # /tmp/foo.m3u as being inside /tmp/foo, simply because they have a
        # common prefix, /tmp/foo, which matches the base path, /tmp/foo.
        path = path.parent

    # Check if dir of file is the base path or a subdir
    try:
        path.relative_to(base_path)
    except ValueError:
        return False
    else:
        return True



import logging
import os

import mopidy
from mopidy import config, ext

logger = logging.getLogger(__name__)


class Extension(ext.Extension):

    dist_name = "Mopidy-M3U"
    ext_name = "m3u"
    version = mopidy.__version__

    def get_default_config(self):
        conf_file = os.path.join(os.path.dirname(__file__), "ext.conf")
        return config.read(conf_file)

    def get_config_schema(self):
        schema = super().get_config_schema()
        schema["base_dir"] = config.Path(optional=True)
        schema["default_encoding"] = config.String()
        schema["default_extension"] = config.String(choices=[".m3u", ".m3u8"])
        schema["playlists_dir"] = config.Path(optional=True)
        return schema

    def setup(self, registry):
        from .backend import M3UBackend

        registry.add("backend", M3UBackend)



import pathlib
import platform
import shutil
import tempfile
import unittest

import pykka

from mopidy import core
from mopidy.m3u.backend import M3UBackend
from mopidy.models import Playlist, Track

from tests import dummy_audio, path_to_data_dir
from tests.m3u import generate_song


class M3UPlaylistsProviderTest(unittest.TestCase):
    backend_class = M3UBackend
    config = {
        "m3u": {
            "enabled": True,
            "base_dir": None,
            "default_encoding": "latin-1",
            "default_extension": ".m3u",
            "playlists_dir": path_to_data_dir(""),
        }
    }

    def setUp(self):  # noqa: N802
        self.config["m3u"]["playlists_dir"] = pathlib.Path(tempfile.mkdtemp())
        self.playlists_dir = self.config["m3u"]["playlists_dir"]
        self.base_dir = self.config["m3u"]["base_dir"] or self.playlists_dir

        audio = dummy_audio.create_proxy()
        backend = M3UBackend.start(config=self.config, audio=audio).proxy()
        self.core = core.Core(backends=[backend])

    def tearDown(self):  # noqa: N802
        pykka.ActorRegistry.stop_all()

        if self.playlists_dir.exists():
            shutil.rmtree(str(self.playlists_dir))

    def test_created_playlist_is_persisted(self):
        uri = "m3u:test.m3u"
        path = self.playlists_dir / "test.m3u"
        assert not path.exists()

        playlist = self.core.playlists.create("test")
        assert "test" == playlist.name
        assert uri == playlist.uri
        assert path.exists()

    def test_create_sanitizes_playlist_name(self):
        playlist = self.core.playlists.create("  ../../test FOO baR ")
        assert "..|..|test FOO baR" == playlist.name
        path = self.playlists_dir / "..|..|test FOO baR.m3u"
        assert self.playlists_dir == path.parent
        assert path.exists()

    def test_saved_playlist_is_persisted(self):
        uri1 = "m3u:test1.m3u"
        uri2 = "m3u:test2.m3u"

        path1 = self.playlists_dir / "test1.m3u"
        path2 = self.playlists_dir / "test2.m3u"

        playlist = self.core.playlists.create("test1")
        assert "test1" == playlist.name
        assert uri1 == playlist.uri
        assert path1.exists()
        assert not path2.exists()

        playlist = self.core.playlists.save(playlist.replace(name="test2"))
        assert "test2" == playlist.name
        assert uri2 == playlist.uri
        assert not path1.exists()
        assert path2.exists()

    def test_deleted_playlist_is_removed(self):
        uri = "m3u:test.m3u"
        path = self.playlists_dir / "test.m3u"

        assert not path.exists()

        playlist = self.core.playlists.create("test")
        assert "test" == playlist.name
        assert uri == playlist.uri
        assert path.exists()

        success = self.core.playlists.delete(playlist.uri)
        assert success
        assert not path.exists()

    def test_delete_on_path_outside_playlist_dir_returns_none(self):
        success = self.core.playlists.delete("m3u:///etc/passwd")

        assert not success

    def test_playlist_contents_is_written_to_disk(self):
        track = Track(uri=generate_song(1))
        playlist = self.core.playlists.create("test")
        playlist = self.core.playlists.save(playlist.replace(tracks=[track]))
        path = self.playlists_dir / "test.m3u"

        contents = path.read_text()

        assert track.uri == contents.strip()

    def test_extended_playlist_contents_is_written_to_disk(self):
        track = Track(uri=generate_song(1), name="Test", length=60000)
        playlist = self.core.playlists.create("test")
        playlist = self.core.playlists.save(playlist.replace(tracks=[track]))
        path = self.playlists_dir / "test.m3u"

        m3u = path.read_text().splitlines()

        assert ["#EXTM3U", "#EXTINF:-1,Test", track.uri] == m3u

    def test_latin1_playlist_contents_is_written_to_disk(self):
        track = Track(uri=generate_song(1), name="Test\x9f", length=60000)
        playlist = self.core.playlists.create("test")
        playlist = self.core.playlists.save(playlist.replace(tracks=[track]))
        path = self.playlists_dir / "test.m3u"

        m3u = path.read_bytes().splitlines()

        track_uri = track.uri
        if not isinstance(track_uri, bytes):
            track_uri = track.uri.encode()
        assert [b"#EXTM3U", b"#EXTINF:-1,Test\x9f", track_uri] == m3u

    def test_utf8_playlist_contents_is_replaced_and_written_to_disk(self):
        track = Track(uri=generate_song(1), name="Test\u07b4", length=60000)
        playlist = self.core.playlists.create("test")
        playlist = self.core.playlists.save(playlist.replace(tracks=[track]))
        path = self.playlists_dir / "test.m3u"

        m3u = path.read_bytes().splitlines()

        track_uri = track.uri
        if not isinstance(track_uri, bytes):
            track_uri = track.uri.encode()
        assert [b"#EXTM3U", b"#EXTINF:-1,Test?", track_uri] == m3u

    def test_playlists_are_loaded_at_startup(self):
        track = Track(uri="dummy:track:path2")
        playlist = self.core.playlists.create("test")
        playlist = playlist.replace(tracks=[track])
        playlist = self.core.playlists.save(playlist)

        assert len(self.core.playlists.as_list()) == 1
        result = self.core.playlists.lookup(playlist.uri)
        assert playlist.uri == result.uri
        assert playlist.name == result.name
        assert track.uri == result.tracks[0].uri

    @unittest.skipIf(
        platform.system() == "Darwin",
        'macOS 10.13 raises IOError "Illegal byte sequence" on open.',
    )
    def test_load_playlist_with_nonfilesystem_encoding_of_filename(self):
        playlist_name = "øæå.m3u".encode("latin-1")
        playlist_name = playlist_name.decode(errors="surrogateescape")
        path = self.playlists_dir / playlist_name
        path.write_bytes(b"#EXTM3U\n")

        self.core.playlists.refresh()

        assert len(self.core.playlists.as_list()) == 1
        result = self.core.playlists.as_list()
        assert "���" == result[0].name

    @unittest.SkipTest
    def test_playlists_dir_is_created(self):
        pass

    def test_create_returns_playlist_with_name_set(self):
        playlist = self.core.playlists.create("test")
        assert playlist.name == "test"

    def test_create_returns_playlist_with_uri_set(self):
        playlist = self.core.playlists.create("test")
        assert playlist.uri

    def test_create_adds_playlist_to_playlists_collection(self):
        playlist = self.core.playlists.create("test")
        playlists = self.core.playlists.as_list()
        assert playlist.uri in [ref.uri for ref in playlists]

    def test_as_list_empty_to_start_with(self):
        assert len(self.core.playlists.as_list()) == 0

    def test_as_list_ignores_non_playlists(self):
        path = self.playlists_dir / "test.foo"
        path.touch()
        assert path.exists()

        self.core.playlists.refresh()

        assert len(self.core.playlists.as_list()) == 0

    def test_delete_non_existant_playlist(self):
        self.core.playlists.delete("m3u:unknown")

    def test_delete_playlist_removes_it_from_the_collection(self):
        playlist = self.core.playlists.create("test")
        assert playlist == self.core.playlists.lookup(playlist.uri)

        self.core.playlists.delete(playlist.uri)

        assert self.core.playlists.lookup(playlist.uri) is None

    def test_delete_playlist_without_file(self):
        playlist = self.core.playlists.create("test")
        assert playlist == self.core.playlists.lookup(playlist.uri)

        path = self.playlists_dir / "test.m3u"
        assert path.exists()

        path.unlink()
        assert not path.exists()

        self.core.playlists.delete(playlist.uri)
        assert self.core.playlists.lookup(playlist.uri) is None

    def test_lookup_finds_playlist_by_uri(self):
        original_playlist = self.core.playlists.create("test")

        looked_up_playlist = self.core.playlists.lookup(original_playlist.uri)

        assert original_playlist == looked_up_playlist

    def test_lookup_on_path_outside_playlist_dir_returns_none(self):
        result = self.core.playlists.lookup("m3u:///etc/passwd")

        assert result is None

    def test_refresh(self):
        playlist = self.core.playlists.create("test")
        assert playlist == self.core.playlists.lookup(playlist.uri)

        self.core.playlists.refresh()

        assert playlist == self.core.playlists.lookup(playlist.uri)

    def test_save_replaces_existing_playlist_with_updated_playlist(self):
        playlist1 = self.core.playlists.create("test1")
        assert playlist1 == self.core.playlists.lookup(playlist1.uri)

        playlist2 = playlist1.replace(name="test2")
        playlist2 = self.core.playlists.save(playlist2)
        assert self.core.playlists.lookup(playlist1.uri) is None
        assert playlist2 == self.core.playlists.lookup(playlist2.uri)

    def test_create_replaces_existing_playlist_with_updated_playlist(self):
        track = Track(uri=generate_song(1))
        playlist1 = self.core.playlists.create("test")
        playlist1 = self.core.playlists.save(playlist1.replace(tracks=[track]))
        assert playlist1 == self.core.playlists.lookup(playlist1.uri)

        playlist2 = self.core.playlists.create("test")
        assert playlist1.uri == playlist2.uri
        assert playlist1 != self.core.playlists.lookup(playlist1.uri)
        assert playlist2 == self.core.playlists.lookup(playlist1.uri)

    def test_save_playlist_with_new_uri(self):
        uri = "m3u:test.m3u"
        self.core.playlists.save(Playlist(uri=uri))
        path = self.playlists_dir / "test.m3u"
        assert path.exists()

    def test_save_on_path_outside_playlist_dir_returns_none(self):
        result = self.core.playlists.save(Playlist(uri="m3u:///tmp/test.m3u"))
        assert result is None

    def test_playlist_with_unknown_track(self):
        track = Track(uri="file:///dev/null")
        playlist = self.core.playlists.create("test")
        playlist = playlist.replace(tracks=[track])
        playlist = self.core.playlists.save(playlist)

        assert len(self.core.playlists.as_list()) == 1
        result = self.core.playlists.lookup("m3u:test.m3u")
        assert "m3u:test.m3u" == result.uri
        assert playlist.name == result.name
        assert track.uri == result.tracks[0].uri

    def test_playlist_with_absolute_path(self):
        track = Track(uri="/tmp/test.mp3")
        filepath = pathlib.Path("/tmp/test.mp3")
        playlist = self.core.playlists.create("test")
        playlist = playlist.replace(tracks=[track])
        playlist = self.core.playlists.save(playlist)

        assert len(self.core.playlists.as_list()) == 1
        result = self.core.playlists.lookup("m3u:test.m3u")
        assert "m3u:test.m3u" == result.uri
        assert playlist.name == result.name
        assert filepath.as_uri() == result.tracks[0].uri

    def test_playlist_with_relative_path(self):
        track = Track(uri="test.mp3")
        filepath = self.base_dir / "test.mp3"
        playlist = self.core.playlists.create("test")
        playlist = playlist.replace(tracks=[track])
        playlist = self.core.playlists.save(playlist)

        assert len(self.core.playlists.as_list()) == 1
        result = self.core.playlists.lookup("m3u:test.m3u")
        assert "m3u:test.m3u" == result.uri
        assert playlist.name == result.name
        assert filepath.as_uri() == result.tracks[0].uri

    def test_playlist_sort_order(self):
        def check_order(playlists, names):
            assert names == [playlist.name for playlist in playlists]

        self.core.playlists.create("c")
        self.core.playlists.create("a")
        self.core.playlists.create("b")

        check_order(self.core.playlists.as_list(), ["a", "b", "c"])

        self.core.playlists.refresh()

        check_order(self.core.playlists.as_list(), ["a", "b", "c"])

        playlist = self.core.playlists.lookup("m3u:a.m3u")
        playlist = playlist.replace(name="d")
        playlist = self.core.playlists.save(playlist)

        check_order(self.core.playlists.as_list(), ["b", "c", "d"])

        self.core.playlists.delete("m3u:c.m3u")

        check_order(self.core.playlists.as_list(), ["b", "d"])

    def test_get_items_returns_item_refs(self):
        track = Track(uri="dummy:a", name="A", length=60000)
        playlist = self.core.playlists.create("test")
        playlist = self.core.playlists.save(playlist.replace(tracks=[track]))

        item_refs = self.core.playlists.get_items(playlist.uri)

        assert len(item_refs) == 1
        assert item_refs[0].type == "track"
        assert item_refs[0].uri == "dummy:a"
        assert item_refs[0].name == "A"

    def test_get_items_of_unknown_playlist_returns_none(self):
        item_refs = self.core.playlists.get_items("dummy:unknown")

        assert item_refs is None

    def test_get_items_from_file_outside_playlist_dir_returns_none(self):
        item_refs = self.core.playlists.get_items("m3u:///etc/passwd")

        assert item_refs is None


class M3UPlaylistsProviderBaseDirectoryTest(M3UPlaylistsProviderTest):
    def setUp(self):  # noqa: N802
        self.config["m3u"]["base_dir"] = pathlib.Path(tempfile.mkdtemp())
        super().setUp()



import urllib
from collections.abc import Iterable, Mapping

from mopidy import exceptions

PLAYBACK_STATES = {"paused", "stopped", "playing"}

SEARCH_FIELDS = {
    "uri",
    "track_name",
    "album",
    "artist",
    "albumartist",
    "composer",
    "performer",
    "track_no",
    "genre",
    "date",
    "comment",
    "any",
}

PLAYLIST_FIELDS = {"uri", "name"}  # TODO: add length and last_modified?

TRACKLIST_FIELDS = {  # TODO: add bitrate, length, disc_no, track_no, modified?
    "uri",
    "name",
    "genre",
    "date",
    "comment",
    "musicbrainz_id",
}

DISTINCT_FIELDS = {
    "track",
    "artist",
    "albumartist",
    "album",
    "composer",
    "performer",
    "date",
    "genre",
}


# TODO: _check_iterable(check, msg, **kwargs) + [check(a) for a in arg]?
def _check_iterable(arg, msg, **kwargs):
    """Ensure we have an iterable which is not a string or an iterator"""
    if isinstance(arg, str):
        raise exceptions.ValidationError(msg.format(arg=arg, **kwargs))
    elif not isinstance(arg, Iterable):
        raise exceptions.ValidationError(msg.format(arg=arg, **kwargs))
    elif iter(arg) is iter(arg):
        raise exceptions.ValidationError(msg.format(arg=arg, **kwargs))


def check_choice(arg, choices, msg="Expected one of {choices}, not {arg!r}"):
    if arg not in choices:
        raise exceptions.ValidationError(
            msg.format(arg=arg, choices=tuple(choices))
        )


def check_boolean(arg, msg="Expected a boolean, not {arg!r}"):
    check_instance(arg, bool, msg=msg)


def check_instance(arg, cls, msg="Expected a {name} instance, not {arg!r}"):
    if not isinstance(arg, cls):
        raise exceptions.ValidationError(msg.format(arg=arg, name=cls.__name__))


def check_instances(arg, cls, msg="Expected a list of {name}, not {arg!r}"):
    _check_iterable(arg, msg, name=cls.__name__)
    if not all(isinstance(instance, cls) for instance in arg):
        raise exceptions.ValidationError(msg.format(arg=arg, name=cls.__name__))


def check_integer(arg, min=None, max=None):
    if not isinstance(arg, int):
        raise exceptions.ValidationError(f"Expected an integer, not {arg!r}")
    elif min is not None and arg < min:
        raise exceptions.ValidationError(
            f"Expected number larger or equal to {min}, not {arg!r}"
        )
    elif max is not None and arg > max:
        raise exceptions.ValidationError(
            f"Expected number smaller or equal to {max}, not {arg!r}"
        )


def check_query(arg, fields=None, list_values=True):
    if fields is None:
        fields = SEARCH_FIELDS
    # TODO: normalize name  -> track_name
    # TODO: normalize value -> [value]
    # TODO: normalize blank -> [] or just remove field?
    # TODO: remove list_values?

    if not isinstance(arg, Mapping):
        raise exceptions.ValidationError(
            f"Expected a query dictionary, not {arg!r}"
        )

    for key, value in arg.items():
        check_choice(
            key,
            fields,
            msg="Expected query field to be one of " "{choices}, not {arg!r}",
        )
        if list_values:
            msg = 'Expected "{key}" to be list of strings, not {arg!r}'
            _check_iterable(value, msg, key=key)
            [_check_query_value(key, v, msg) for v in value]
        else:
            _check_query_value(
                key, value, 'Expected "{key}" to be a string, not {arg!r}'
            )


def _check_query_value(key, arg, msg):
    if not isinstance(arg, str) or not arg.strip():
        raise exceptions.ValidationError(msg.format(arg=arg, key=key))


def check_uri(arg, msg="Expected a valid URI, not {arg!r}"):
    if not isinstance(arg, str):
        raise exceptions.ValidationError(msg.format(arg=arg))
    elif urllib.parse.urlparse(arg).scheme == "":
        raise exceptions.ValidationError(msg.format(arg=arg))


def check_uris(arg, msg="Expected a list of URIs, not {arg!r}"):
    _check_iterable(arg, msg)
    [check_uri(a, msg) for a in arg]



import logging

import pykka

from mopidy import backend
from mopidy.file import library

logger = logging.getLogger(__name__)


class FileBackend(pykka.ThreadingActor, backend.Backend):
    uri_schemes = ["file"]

    def __init__(self, config, audio):
        super().__init__()
        self.library = library.FileLibraryProvider(backend=self, config=config)
        self.playback = backend.PlaybackProvider(audio=audio, backend=self)
        self.playlists = None



import os
import unittest
from unittest import mock

import tornado.testing
import tornado.web
import tornado.websocket

import mopidy
from mopidy.http import handlers


class StaticFileHandlerTest(tornado.testing.AsyncHTTPTestCase):
    def get_app(self):
        return tornado.web.Application(
            [
                (
                    r"/(.*)",
                    handlers.StaticFileHandler,
                    {
                        "path": os.path.dirname(__file__),
                        "default_filename": "test_handlers.py",
                    },
                )
            ]
        )

    def test_static_handler(self):
        response = self.fetch("/test_handlers.py", method="GET")

        assert 200 == response.code
        assert response.headers["X-Mopidy-Version"] == mopidy.__version__
        assert response.headers["Cache-Control"] == "no-cache"

    def test_static_default_filename(self):
        response = self.fetch("/", method="GET")

        assert 200 == response.code
        assert response.headers["X-Mopidy-Version"] == mopidy.__version__
        assert response.headers["Cache-Control"] == "no-cache"


class WebSocketHandlerTest(tornado.testing.AsyncHTTPTestCase):
    def get_app(self):
        self.core = mock.Mock()
        return tornado.web.Application(
            [
                (
                    r"/ws/?",
                    handlers.WebSocketHandler,
                    {
                        "core": self.core,
                        "allowed_origins": [],
                        "csrf_protection": True,
                    },
                )
            ]
        )

    def connection(self):
        url = self.get_url("/ws").replace("http", "ws")
        return tornado.websocket.websocket_connect(url)

    @tornado.testing.gen_test
    def test_invalid_json_rpc_request_doesnt_crash_handler(self):
        # An uncaught error would result in no message, so this is just a
        # simplistic test to verify this.
        conn = yield self.connection()
        conn.write_message("invalid request")
        message = yield conn.read_message()
        assert message

    @tornado.testing.gen_test
    def test_broadcast_makes_it_to_client(self):
        conn = yield self.connection()
        handlers.WebSocketHandler.broadcast("message", self.io_loop)
        message = yield conn.read_message()
        assert message == "message"

    @tornado.testing.gen_test
    def test_broadcast_to_client_that_just_closed_connection(self):
        conn = yield self.connection()
        conn.stream.close()
        handlers.WebSocketHandler.broadcast("message", self.io_loop)

    @tornado.testing.gen_test
    def test_broadcast_to_client_without_ws_connection_present(self):
        yield self.connection()
        # Tornado checks for ws_connection and raises WebSocketClosedError
        # if it is missing, this test case simulates winning a race were
        # this has happened but we have not yet been removed from clients.
        for client in handlers.WebSocketHandler.clients:
            client.ws_connection = None
        handlers.WebSocketHandler.broadcast("message", self.io_loop)


class CheckOriginTests(unittest.TestCase):
    def setUp(self):
        self.headers = {"Host": "localhost:6680"}
        self.allowed = set()

    def test_missing_origin_blocked(self):
        assert not handlers.check_origin(None, self.headers, self.allowed)

    def test_empty_origin_allowed(self):
        assert handlers.check_origin("", self.headers, self.allowed)

    def test_chrome_file_origin_allowed(self):
        assert handlers.check_origin("file://", self.headers, self.allowed)

    def test_firefox_null_origin_allowed(self):
        assert handlers.check_origin("null", self.headers, self.allowed)

    def test_same_host_origin_allowed(self):
        assert handlers.check_origin(
            "http://localhost:6680", self.headers, self.allowed
        )

    def test_different_host_origin_blocked(self):
        assert not handlers.check_origin(
            "http://other:6680", self.headers, self.allowed
        )

    def test_different_port_blocked(self):
        assert not handlers.check_origin(
            "http://localhost:80", self.headers, self.allowed
        )

    def test_extra_origin_allowed(self):
        self.allowed.add("other:6680")
        assert handlers.check_origin(
            "http://other:6680", self.headers, self.allowed
        )
